import numpy as np
from numba import jit
from ..tinyhouse import InputOutput


def readInSeg(pedigree, fileName, start=None, stop = None):
    print("Reading in seg file:", fileName)
    if start is None: start = 0
    if stop is None: stop = pedigree.nLoci
    nLoci = stop - start + 1 #Contains stop.
    
    seg = np.full((pedigree.maxIdn, 4, nLoci), .25, dtype = np.float32)

    index = 0
    fileNColumns = 0

    indHit = np.full(pedigree.maxIdn, 0, dtype = np.int64)

    with open(fileName) as f:
        e = 0
        currentInd = None
        for line in f:
            parts = line.split();
            idx = parts[0]; 

            if fileNColumns == 0:
                fileNColumns = len(parts)
            if fileNColumns != len(parts):
                raise ValueError(f"The length of the line is not the expected length. Expected {fileNColumns} got {len(parts)} on individual {idx} and line {e}.")

            segLine=np.array([float(val) for val in parts[(start+1):(stop+2)]], dtype = np.float32)
            if len(segLine) != nLoci:
                raise ValueError(f"The length of the line subsection is not the expected length. Expected {nLoci} got {len(segLine)} on individual {idx} and line {e}.")

            if idx not in pedigree.individuals:
                print(f"Individual {idx} not found in pedigree. Individual ignored.")
            else:
                ind = pedigree.individuals[idx]
                if e == 0: 
                    currentInd = ind.idx
                if currentInd != ind.idx:
                    raise ValueError(f"Unexpected individual. Expecting individual {currentInd}, but got ind {ind.idx} on value {e}")
                seg[ind.idn,e,:] = segLine
                e = (e+1)%4
                ind.fileIndex['segregation'] = index; index += 1
                indHit[ind.idn] += 1
        for ind in pedigree:
            if indHit[ind.idn] != 4:
                print(f"No segregation information found for individual {ind.idx}")

    return seg

def writeOutParamaters(peelingInfo) :
    args = InputOutput.args

    np.savetxt(args.out + ".genoError", peelingInfo.genoError, fmt = "%f")
    np.savetxt(args.out + ".seqError", peelingInfo.seqError, fmt = "%f")
    # np.savetxt(args.out + ".trans", peelingInfo.transmissionRate, fmt = "%f")
    np.savetxt(args.out + ".maf", peelingInfo.maf, fmt = "%f")

def writeGenotypes(pedigree, genoProbFunc) :
    args = InputOutput.args
    if not args.no_dosages: writeDosages(pedigree, genoProbFunc, args.out + ".dosages")
    if args.haps: writeGenoProbs(pedigree, genoProbFunc, args.out + ".haps")

    if args.calling_threshold is not None:
        for thresh in args.calling_threshold:
            if args.binary_call_files : writeBinaryCalledGenotypes(pedigree, genoProbFunc, args.out + ".called." + str(thresh), thresh)
            if not args.binary_call_files : writeCalledGenotypes(pedigree, genoProbFunc, args.out + ".called." + str(thresh), thresh)

            if args.call_phase:
                writeCalledPhase(pedigree, genoProbFunc, args.out + ".called_phase." + str(thresh), thresh)

def writeGenoProbs(pedigree, genoProbFunc, outputFile):
    with open(outputFile, 'w+') as f:
        for idx, ind in pedigree.writeOrder():
            matrix = genoProbFunc(ind.idn)
            f.write('\n')
            for i in range(matrix.shape[0]) :
                f.write(ind.idx + ' ' + ' '.join(map("{:.4f}".format, matrix[i,:])) + '\n')

def writeDosages(pedigree, genoProbFunc, outputFile):
    with open(outputFile, 'w+') as f:
        for idx, ind in pedigree.writeOrder():
            matrix = np.dot(np.array([0,1,1,2]), genoProbFunc(ind.idn))
            
            if InputOutput.args.sexchrom and ind.sex == 0:
                matrix *= 2

            f.write(ind.idx + ' ' + ' '.join(map("{:.4f}".format, matrix)) + '\n')


def writeCalledGenotypes(pedigree, genoProbFunc, outputFile, thresh):
    with open(outputFile, 'w+') as f:
        for idx, ind in pedigree.writeOrder():
            matrix = genoProbFunc(ind.idn)
            
            matrixCollapsedHets = np.array([matrix[0,:], matrix[1,:] + matrix[2,:], matrix[3,:]], dtype=np.float32)
            calledGenotypes = np.argmax(matrixCollapsedHets, axis = 0)
            setMissing(calledGenotypes, matrixCollapsedHets, thresh)
            if InputOutput.args.sexchrom and ind.sex == 0:
                doubleIfNotMissing(calledGenotypes)

            f.write(ind.idx + ' ' + ' '.join(map(str, calledGenotypes)) + '\n')

def writeCalledPhase(pedigree, genoProbFunc, outputFile, thresh):
    with open(outputFile, 'w+') as f:
        for idx, ind in pedigree.writeOrder():
            matrix = genoProbFunc(ind.idn)

            # Paternal            
            paternal_probs = np.array([matrix[0,:] + matrix[1,:], matrix[2,:] + matrix[3,:]], dtype=np.float32)
            paternal_haplotype = np.argmax(paternal_probs, axis = 0)
            setMissing(paternal_haplotype, paternal_probs, thresh)
            f.write(ind.idx + ' ' + ' '.join(map(str, paternal_haplotype)) + '\n')

            #Maternal
            maternal_probs = np.array([matrix[0,:] + matrix[2,:], matrix[1,:] + matrix[3,:]], dtype=np.float32)
            maternal_haplotype = np.argmax(maternal_probs, axis = 0)
            setMissing(maternal_haplotype, maternal_probs, thresh)
            f.write(ind.idx + ' ' + ' '.join(map(str, maternal_haplotype)) + '\n')

def writeBinaryCalledGenotypes(pedigree, genoProbFunc, outputFile, thresh):
    for idx, ind in pedigree.writeOrder():
        matrix = genoProbFunc(ind.idn)
        matrixCollapsedHets = np.array([matrix[0,:], matrix[1,:] + matrix[2,:], matrix[3,:]], dtype=np.float32)
        calledGenotypes = np.argmax(matrixCollapsedHets, axis = 0)
        setMissing(calledGenotypes, matrixCollapsedHets, thresh)
        if InputOutput.args.sexchrom and ind.sex == 0:
            doubleIfNotMissing(calledGenotypes)
        ind.genotypes = calledGenotypes.astype(np.int8)

    InputOutput.writeOutGenotypesPlink(pedigree, outputFile)

@jit(nopython=True)
def doubleIfNotMissing(calledGenotypes):
    nLoci = len(calledGenotypes)
    for i in range(nLoci):
        if calledGenotypes[i] == 1:
            calledGenotypes[i] = 2

@jit(nopython=True)
def setMissing(calledGenotypes, matrix, thresh) :
    nLoci = len(calledGenotypes)
    for i in range(nLoci):
        if matrix[calledGenotypes[i],i] < thresh:
            calledGenotypes[i] = 9

def fullOutput(pedigree, peelingInfo, args):
    InputOutput.writeIdnIndexedMatrix(pedigree, peelingInfo.penetrance, args.out + ".penetrance")
    InputOutput.writeIdnIndexedMatrix(pedigree, peelingInfo.anterior, args.out + ".anterior")
    InputOutput.writeIdnIndexedMatrix(pedigree, peelingInfo.posterior, args.out + ".posterior")

    InputOutput.writeFamIndexedMatrix(pedigree, peelingInfo.posteriorSire_minusFam, args.out + ".posteriorSire_minusFam")
    InputOutput.writeFamIndexedMatrix(pedigree, peelingInfo.posteriorDam_minusFam, args.out + ".posteriorDam_minusFam")

    InputOutput.writeFamIndexedMatrix(pedigree, peelingInfo.posteriorSire_new, args.out + ".posteriorSire_new")
    InputOutput.writeFamIndexedMatrix(pedigree, peelingInfo.posteriorDam_new, args.out + ".posteriorDam_new")



import concurrent.futures
from numba import jit, float32, int8, int64, optional, boolean
from numba.experimental import jitclass
import numpy as np
from collections import OrderedDict

from ..tinyhouse import InputOutput
from ..tinyhouse import ProbMath
from ..tinyhouse import HaplotypeOperations

import math

#########################################################################
### In this module we define the peeling info object.                 ###  
### This is a just in time container for the various                  ### 
### peeling probability calculations.                                 ### 
#########################################################################

def createPeelingInfo(pedigree, args, createSeg=True, phaseFounder = False) :

    # NOTE: createSeg is added as an option to decrease memory usage during the single locus peeling steps.
    nLoci = pedigree.nLoci

    peelingInfo = jit_peelingInformation(nInd=pedigree.maxIdn, nFam=pedigree.maxFam, nLoci=nLoci, createSeg=createSeg)
    peelingInfo.isSexChrom = args.sexchrom
    # Information about the peeling positions are handled elsewhere.
    peelingInfo.positions = None 

    #Generate the segregation tensors. 
    peelingInfo.segregationTensor = ProbMath.generateSegregation(e = 1e-06)
    peelingInfo.segregationTensor_norm = ProbMath.generateSegregation(e = 1e-06, partial=True) #Partial gives the normalizing constant.

    peelingInfo.segregationTensorXY = ProbMath.generateSegregationXYChrom(e = 1e-06)
    peelingInfo.segregationTensorXX = ProbMath.generateSegregationXXChrom(e = 1e-06)

    peelingInfo.genoError[:] = args.error
    peelingInfo.seqError[:] = args.seqerror
    setupTransmission(args.length, peelingInfo) #Sets up the transmission rates using a custom position list and a total chromosome length.

    for ind in pedigree:
        peelingInfo.sex[ind.idn] = ind.sex

        if ind.genotypes is not None and ind.haplotypes is not None:
            HaplotypeOperations.ind_fillInGenotypesFromPhase(ind)

        sexChromFlag = peelingInfo.isSexChrom and ind.sex == 0 #This is the sex chromosome and the individual is male.

        peelingInfo.penetrance[ind.idn,:,:] = ProbMath.getGenotypeProbabilities(peelingInfo.nLoci, ind.genotypes, ind.reads, peelingInfo.genoError, peelingInfo.seqError, sexChromFlag)
        
        # Set the genotyping/read status for each individual. This will be used for, e.g., estimating the minor allele frequency.
        if ind.genotypes is not None:
            setGenotypeStatusGenotypes(ind.idn, ind.genotypes, peelingInfo)

        if ind.reads is not None:
            setGenotypeStatusReads(ind.idn, ind.reads[0], ind.reads[1], peelingInfo)

        if ind.isGenotypedFounder() and phaseFounder and ind.genotypes is not None:
            loci = getHetMidpoint(ind.genotypes)
            if loci is not None:
                e = args.error
                peelingInfo.penetrance[ind.idn,:,loci] = np.array([e/3, e/3, 1-e, e/3], dtype = np.float32)
    

    if args.penetrance is not None:
        if args.sexchrom:
            print("Using an external penetrance file and the sexchrom option is highly discouraged. Please do not use.")

        if args.esterrors :
            print("External penetrance file included, but esterrors flag used. The two options are incompatible. esterrors set to false.")
            args.esterrors = False
        for pen in args.penetrance:
            addPenetranceFromExternalFile(pedigree, peelingInfo, pen, args)
    # updateMaf(pedigree, peelingInfo)
    return peelingInfo

def setupTransmission(length, peelingInfo) :
    if peelingInfo.positions is None:
        localMap = np.linspace(0, 1, num = peelingInfo.nLoci, dtype = np.float32)
    else:
        localMap = peelingInfo.positions/peelingInfo.positions[-1] #This should be sorted. Need to add in code to check.
    for i in range(peelingInfo.nLoci -1):
        distance = localMap[i+1] - localMap[i]
        distance = distance * length
        peelingInfo.transmissionRate[i] = distance



@jit(nopython=True)
def setGenotypeStatusGenotypes(idn, genotypes, peelingInfo):
    nLoci = len(genotypes)
    if genotypes is not None:
        for i in range(nLoci):
            peelingInfo.genotyped[idn, i] = peelingInfo.genotyped[idn, i] or genotypes[i] != 9

@jit(nopython=True)
def setGenotypeStatusReads(idn, reads0, reads1, peelingInfo):
    nLoci = len(reads0)
    if reads0 is not None and reads1 is not None:
        for i in range(nLoci):
            peelingInfo.genotyped[idn, i] = peelingInfo.genotyped[idn, i] or reads0[i] != 0 or reads1[i] != 0

def addPenetranceFromExternalFile(pedigree, peelingInfo, fileName, args):
    # This function allows external penetrance files to be read in and added to the gentoype probabilities for an individual.

    print("Reading in penetrance file:", fileName)
    with open(fileName) as f:
        e = 0
        for line in f:
            parts = line.split(); 
            idx = parts[0]; 
            parts = parts[1:]

            if args.startsnp is not None :
                parts = parts[args.startsnp : args.stopsnp+1] #Offset 1 to include stopsnp

            penetranceLine=np.array([float(val) for val in parts], dtype = np.float32)

            if idx not in pedigree.individuals:
                print("Individual", idx, "not found in pedigree. Individual ignored.")
            else:
                ind = pedigree.individuals[idx]
                peelingInfo.penetrance[ind.idn,e,:] *= penetranceLine
                # Normalizing in terms of SNPs seems like a really bad idea.
                # peelingInfo.penetrance[ind.idn,e,:] /= np.sum(peelingInfo.penetrance[ind.idn,e,:], 0) # Normalization added, just in case.
                e = (e+1)%4


@jit(nopython=True)
def getHetMidpoint(geno):
    nLoci = len(geno)
    midpoint = int(nLoci/2)
    index = 0
    changed = False
    while index < nLoci/2:
        if midpoint + index < nLoci:
            if geno[midpoint + index] == 1:
                return midpoint + index
        if midpoint - index >= 0:
            if geno[midpoint - index] == 1:
                return midpoint - index
        index += 1
    return None

spec = OrderedDict()
spec['nInd'] = int64
spec['nFam'] = int64
spec['nLoci'] = int64

spec['isSexChrom'] = boolean
spec['sex'] = int64[:]
spec['genotyped'] = boolean[:,:] #Maybe this should be removed?

# Individual terms: Each will be nInd x 4 x nLoci
spec['anterior'] = float32[:,:,:]
spec['posterior'] = float32[:,:,:]
spec['penetrance'] = float32[:,:,:]
spec['segregation'] = optional(float32[:,:,:])
spec['pointSeg'] = optional(float32[:,:,:]) # I think we don't use this any more. Potentially could be dropped.

# Family terms. Each will be nFam x 4 x nLoci
spec['posteriorSire_minusFam'] = float32[:,:,:]
spec['posteriorDam_minusFam'] = float32[:,:,:]
spec['posteriorSire_new'] = float32[:,:,:]
spec['posteriorDam_new'] = float32[:,:,:]

# Segregation tensors. Each of these will be either 4x4x4x4 or 4x4x4
spec['segregationTensor'] = optional(float32[:,:,:,:])
spec['segregationTensor_norm'] = optional(float32[:,:,:]) # Note: This one is a bit smaller.
spec['segregationTensorXX'] = optional(float32[:,:,:,:])
spec['segregationTensorXY'] = optional(float32[:,:,:,:])

# Marker specific rates:
spec['genoError'] = optional(float32[:])
spec['seqError'] = optional(float32[:])
spec['transmissionRate'] = optional(float32[:])
spec['maf'] = optional(float32[:])

spec['positions'] = optional(float32[:]) # Not sure we use this.
spec['iteration'] = int64

@jitclass(spec)
class jit_peelingInformation(object):
    def __init__(self, nInd, nFam, nLoci, createSeg=True):
        self.iteration = 0
        self.nInd = nInd
        self.nFam = nFam
        self.nLoci = nLoci
        
        self.isSexChrom = False

        self.construct(createSeg)

        # These are filled in from createPeelingInfo, above.
        self.positions = None
        self.segregationTensor = None
        self.segregationTensor_norm = None

    def construct(self, createSeg = True) :
        baseValue = .25
        self.sex = np.full(self.nInd, 0, dtype = np.int64)

        self.genotyped = np.full((self.nInd, self.nLoci), False, dtype = np.bool_)

        self.anterior = np.full((self.nInd, 4, self.nLoci), baseValue, dtype = np.float32)
        self.posterior = np.full((self.nInd, 4, self.nLoci), baseValue, dtype = np.float32)
        self.penetrance = np.full((self.nInd, 4, self.nLoci), baseValue, dtype = np.float32)
        
        self.segregation = np.full((self.nInd, 4, self.nLoci), baseValue, dtype = np.float32)

        if createSeg: # Only removes the point seg term since this is not used for single locus peeling.
            self.pointSeg = np.full((self.nInd, 4, self.nLoci), baseValue, dtype = np.float32)
        else:
            self.pointSeg = None

        self.posteriorSire_minusFam = np.full((self.nFam, 4, self.nLoci), baseValue, dtype = np.float32)
        self.posteriorDam_minusFam = np.full((self.nFam, 4, self.nLoci), baseValue, dtype = np.float32)

        self.posteriorSire_new = np.full((self.nFam, 4, self.nLoci), baseValue, dtype = np.float32)
        self.posteriorDam_new = np.full((self.nFam, 4, self.nLoci), baseValue, dtype = np.float32)

        self.genoError = np.full((self.nLoci), 0, dtype = np.float32)
        self.seqError = np.full((self.nLoci), 0, dtype = np.float32)
        self.maf = np.full((self.nLoci), .5, dtype = np.float32)
        self.transmissionRate = np.full((self.nLoci-1), 0, dtype = np.float32)


    def getGenoProbs(self, idn):
        genoProbs = self.anterior[idn,:,:]*self.posterior[idn,:,:]*self.penetrance[idn,:,:]
        genoProbs = genoProbs/np.sum(genoProbs,0) 
        return genoProbs





import concurrent.futures
from numba import jit, float32, int8, int64, optional, boolean
from numba.experimental import jitclass
import numpy as np
from collections import OrderedDict

from ..tinyhouse import InputOutput
from ..tinyhouse import ProbMath
from ..tinyhouse import HaplotypeOperations

from . import PeelingInfo

import math

#########################################################################
### In this module we will update 3 things:                           ###  
### 1) Our estimate for the MAF.                                      ### 
### 2) Our estimate of the locus specific (sequencing) error rate.    ###
### 3) Our estimate of the locus specific recombination rate.         ###
#########################################################################


# Estimating the minor allele frequency. This update is done by using an iterative approach 
# which minimizes the likelihood of the observed genotypes conditional on them having been 
# generated from hardy-weinberg equilibrium with a fixed maf value. To speed up, we use
# Newton style updates to re-estimate the minor allele frequency.
# There is math on how to do this... somewhere?

def updateMaf(pedigree, peelingInfo):
    if peelingInfo.isSexChrom:
        print("Updating error rates and minor allele frequencies for sex chromosomes are not well test and will break in interesting ways. Recommend running without that option.")

    maf = np.full(peelingInfo.nLoci, .5, dtype = np.float32)
    for i in range(peelingInfo.nLoci):
        maf[i] = newtonMafUpdates(peelingInfo, i)

    mafGeno = ProbMath.getGenotypesFromMaf(maf)
    for ind in pedigree:
        if ind.isFounder():
            peelingInfo.anterior[ind.idn,:,:] = mafGeno
    peelingInfo.maf = maf.astype(np.float32)

def newtonMafUpdates(peelingInfo, index) :
    # This function gives an iterative approximation for the minor allele frequency. It uses a maximum of 5 iterations.
    maf = 0.5
    maf_old = 0.5
    iters = 5
    converged = False
    while not converged:
        delta = getNewtonUpdate(maf_old, peelingInfo, index)
        maf = maf_old + delta
        if maf < 0.01:
            maf = 0.01
        if maf > .99:
            maf = .99
        if abs(maf - maf_old) < 0.0001:
            converged = True
        iters -= 1
        if iters < 0:
            converged = True
    return maf

@jit(nopython=True)
def getNewtonUpdate(p, peelingInfo, index) :
    #Log likelihood of the liklihood's first + second derivitives
    LLp = 0
    LLpp = 0

    #I want to add priors. Should be 1 individual of each of the four states.
    LLp, LLpp = addIndividualToUpdate(np.array([1, 0, 0, 0], dtype = np.float32), p, LLp, LLpp)
    LLp, LLpp = addIndividualToUpdate(np.array([0, 1, 0, 0], dtype = np.float32), p, LLp, LLpp)
    LLp, LLpp = addIndividualToUpdate(np.array([0, 0, 1, 0], dtype = np.float32), p, LLp, LLpp)
    LLp, LLpp = addIndividualToUpdate(np.array([0, 0, 0, 1], dtype = np.float32), p, LLp, LLpp)
    for i in range(peelingInfo.nInd):
        if peelingInfo.genotyped[i, index] :
            d = peelingInfo.penetrance[i,:,index]
            LLp, LLpp = addIndividualToUpdate(d, p, LLp, LLpp)
    if LLp == 0 or LLpp == 0: return 0 #Could be a case where no one has data.
    return -LLp/LLpp

@jit(nopython = True)
def addIndividualToUpdate(d, p, LLp, LLpp):
    d0 = d[0]
    d1 = d[1] + d[2]
    d2 = d[3]

    f = d0*(1-p)**2 + d1*p*(1-p) + d2*p**2
    fp = (d1 - 2*d0) + 2*p*(d0 + d2 - d1)
    fpp = 2*(d0 + d2 - d1)
    
    LLp += fp/f
    LLpp += fpp/f - (fp/f)**2

    return LLp, LLpp

# Commenting out the following code. This was used to do updates via grid search.
# @jit(nopython = True)
# def mafLoglikelihood(peelingInfo, maf, index):
#     score = 0
#     maf_squared = maf**2
#     maf_one_minus_maf = maf*(1-maf)
#     one_minus_maf_squared = (1-maf)**2

#     for i in range(peelingInfo.nInd):
#         if peelingInfo.genotyped[i, index] :
#         # if True :
#             genoProbs = peelingInfo.penetrance[i,:,index]

#             prob = 0
#             prob += one_minus_maf_squared*genoProbs[0]
#             prob += maf_one_minus_maf*genoProbs[1]
#             prob += maf_one_minus_maf*genoProbs[2]
#             prob += maf_squared*genoProbs[3]
#             score += math.log(prob)
#     return score

###
### NOTE: The following code updates the genotype and sequencing error rates.
###

def updatePenetrance(pedigree, peelingInfo):
    peelingInfo.genoError = updateGenoError(pedigree, peelingInfo)
    peelingInfo.seqError = updateSeqError(pedigree, peelingInfo)

    if peelingInfo.isSexChrom:
        print("Updating error rates and minor allele frequencies for sex chromosomes are not well test and will break in interesting ways. Recommend running without that option.")
    for ind in pedigree:
        sexChromFlag = peelingInfo.isSexChrom and ind.sex == 0 #This is the sex chromosome and the individual is male.
        peelingInfo.penetrance[ind.idn,:,:] = ProbMath.getGenotypeProbabilities(peelingInfo.nLoci, ind.genotypes, ind.reads, peelingInfo.genoError, peelingInfo.seqError, sexChromFlag)

        if ind.isGenotypedFounder() and (not InputOutput.args.nophasefounders) and ind.genotypes is not None:
            loci = PeelingInfo.getHetMidpoint(ind.genotypes)
            if loci is not None:
                e = peelingInfo.genoError[loci]
                peelingInfo.penetrance[ind.idn,:,loci] = np.array([e/3, e/3, 1-e, e/3], dtype = np.float32)

def updateGenoError(pedigree, peelingInfo) :
    # The following is a simple EM update for the genotyping error rate at each locus.
    # This update adds the expected number of errors that an individual marginalizing over their current estimate of their genotype probabilities.
    # We use a max value of 5% and a min value of .0001 percent to make sure the values are reasonable

    counts = np.full(pedigree.nLoci, 1, dtype = np.float32)
    errors = np.full(pedigree.nLoci, 0.01, dtype = np.float32)

    for ind in pedigree:
        updateGenoError_ind(counts, errors, ind.genotypes, peelingInfo.getGenoProbs(ind.idn))
    
    newError = errors/counts 
    newError = np.maximum(np.minimum(newError, .05), .0001)
    return newError

@jit(nopython=True)
def updateGenoError_ind(counts, errors, genotypes, genoProbs):
    for i in range(len(counts)) :
        if genotypes[i] != 9: # Only include non-missing genotypes.
            counts[i] += 1
            if genotypes[i] == 0:
                errors[i] += (genoProbs[1,i] + genoProbs[2, i] + genoProbs[3, i])
            if genotypes[i] == 1:
                errors[i] += (genoProbs[0, i] + genoProbs[3, i])
            if genotypes[i] == 2:
                errors[i] += (genoProbs[0, i] + genoProbs[1,i] + genoProbs[2, i]) 

def updateSeqError(pedigree, peelingInfo) :
    # The following is a simple EM update for the genotyping error rate at each locus.
    # This update adds the expected number of errors that an individual has marginalizing over their current genotype probabilities.
    # This only uses the homozygotic states, heterozygotic states are ignored (in both the counts + errors terms).
    # We use a max value of 5% and a min value of .0001 percent to make sure the values are reasonable

    counts = np.full(pedigree.nLoci, 1, dtype = np.float32)
    errors = np.full(pedigree.nLoci, 0.001, dtype = np.float32)

    for ind in pedigree:
        if ind.reads is not None:
            updateSeqError_ind(counts, errors, ind.reads[0], ind.reads[1], peelingInfo.getGenoProbs(ind.idn))
    
    newError = errors/counts 
    newError = np.maximum(np.minimum(newError, .01), .0001)
    return newError

@jit(nopython=True)
def updateSeqError_ind(counts, errors, refReads, altReads, genoProbs):
    # Errors occur when genotype is 0 and an alternative allele happens.
    # Errors occur when genotype is 2 (coded as 3) and a reference allele happens.
    # Number of observations is number of reads * probability the individual is homozygous.
    for i in range(len(counts)) :
        counts[i] += (genoProbs[0,i] + genoProbs[3,i]) * (altReads[i] + refReads[i])
        errors[i] += genoProbs[0,i] * altReads[i]
        errors[i] += genoProbs[3,i] * refReads[i]


###
### NOTE: The following code is designed to estimate the recombination rate between markers.
### This is currently broken and does not give good estimates. Strongly recommend not using it, and the associated options have been disabled in tinyPeel.py
###




def updateSeg(peelingInfo):
    #Idea: Split the chromosome into ~10 blocks with slightly different starts/stops. Estimate at each one and then sum.
    #Not sure if this is the best way, but probably will work okay.

    distances = [estDistanceFromBreaks(0, peelingInfo.nLoci-1, nBreaks, peelingInfo) for nBreaks in [5, 10, 20]]
    distance = np.mean(distances[0:2])
    print(distances, ":", distance)

    setupTransmission(distance, peelingInfo)

def estDistanceFromBreaks(loc1, loc2, nBreaks, peelingInfo):
    breakPoints = np.floor(np.linspace(loc1, loc2, nBreaks)).astype(dtype = np.int64)
    distances = [getDistance(breakPoints[i], breakPoints[i+1], peelingInfo) for i in range(nBreaks-1)]
    print(nBreaks, ":", distances)
    return sum(distances)

def getDistance(loc1, loc2, peelingInfo):

    patSeg1 = getSumSeg(loc1, peelingInfo)
    patSeg2 = getSumSeg(loc2, peelingInfo)

    patValid = ((patSeg1 > .99) | (patSeg1 < .01)) & ((patSeg2 > .99) | (patSeg2 < .01)) 
    
    # matValid = ((matSeg1 > .99) | (matSeg1 < .01)) & ((matSeg2 > .99) | (matSeg2 < .01)) 
    
    # patRecomb = np.mean(np.abs(patSeg1[patValid] - patSeg2[patValid]))
    # matRecomb = np.mean(np.abs(matSeg1[matValid] - matSeg2[matValid]))
    
    # recomb = (patRecomb + matRecomb)/2
    
    # difference = np.abs(np.round(patSeg1[patValid]) - np.round(patSeg2[patValid]))
    # return np.mean(difference)

    entropy1 = 1 - (-patSeg1*np.log2(patSeg1) - (1-patSeg1)*np.log2(1-patSeg1))
    entropy2 = 1 - (-patSeg2*np.log2(patSeg2) - (1-patSeg2)*np.log2(1-patSeg2))

    difference = patSeg1*(1-patSeg2) + (1-patSeg1)*patSeg2
    est = np.sum(entropy1*entropy2*difference)/np.sum(entropy1*entropy2)
    # return est
    return haldane(est)

def getSumSeg(loc, peelingInfo):
    seg = peelingInfo.segregation[:, :, loc]
    sumSeg = np.sum(seg, 1)
    patSeg = (seg[:,2] + seg[:,3])/sumSeg
    # matSeg = (seg[:,1] + seg[:,3])/sumSeg
    return patSeg


def haldane(difference) :
    return -np.log(1.0-2.0*difference)/2.0




from numba import jit
import numpy as np


@jit(nopython=True, nogil=True)
def multinomial_sample(pvals):
    """Draw one sample from a multinomial distribution (similar to np.random.multinomial)
    pvals   - array of probabilities (np.float32) (that ideally should sum to one - but doesn't have to)
    return  - 1d index of the sampled value. Use multinomial_sample_2d() to convert to a 2d index"""

    # Cumulative sum of the probabilities, total sum is cumsum[-1]
    cumsum = np.cumsum(pvals)
    # Random float in the half-open interval [0.0, sum(pvals)) - this handles pvals that don't sum to 1
    rand = np.random.random() * cumsum[-1]
    # Find where the cumulative sum exceeds the random float and return that index
    for index, value in enumerate(cumsum):
        if value > rand:
            break
    return index


@jit(nopython=True, nogil=True)
def multinomial_sample_2d(pvals):
    """Draw one sample from a multinomial distribution (similar to np.random.multinomial)
    pvals   - 2d array of probabilities (np.float32) (that ideally should sum to one - but doesn't have to)
    return  - (row_index, col_index) tuple"""

    # Get 1d index
    index = multinomial_sample(pvals)
    # Convert 1d index to 2d and return indices tuple
    ncols = pvals.shape[1]
    return (index//ncols, index%ncols)



import numpy as np
from numba import jit, float32, int32, int64, optional
from numba.experimental import jitclass

from .tinyhouse import Pedigree
from .tinyhouse import ProbMath
from .tinyhouse import InputOutput 

from .Peeling import Peeling
from .Peeling import PeelingIO
from .Peeling import PeelingInfo
from .Peeling import PeelingUpdates

import concurrent.futures
from itertools import repeat
import argparse

def runPeelingCycles(pedigree, peelingInfo, args, singleLocusMode = False):
    #Right now maf _only_ uses the penetrance so can be estimated once.
    if args.estmaf: PeelingUpdates.updateMaf(pedigree, peelingInfo)

    for i in range(args.ncycles):
        print("Cycle ", i)
        peelingCycle(pedigree, peelingInfo, args = args, singleLocusMode = singleLocusMode)
        peelingInfo.iteration += 1
        
        # esttransitions is been disabled.
        # if args.esttransitions: 
        #     print("Estimating the transmission rate is currently a disabled option")
            # PeelingUpdates.updateSeg(peelingInfo) #Option currently disabled.
            
        if args.esterrors: 
            PeelingUpdates.updatePenetrance(pedigree, peelingInfo)

def peelingCycle(pedigree, peelingInfo, args, singleLocusMode = False) :
    nWorkers = args.maxthreads
    
    for index, generation in enumerate(pedigree.generations):
        print("Peeling Down, Generation", index)
        jit_families = [family.toJit() for family in generation.families]

        if args.maxthreads > 1:
            with concurrent.futures.ThreadPoolExecutor(max_workers=nWorkers) as executor:
                 results = executor.map(Peeling.peel, jit_families, repeat(Peeling.PEEL_DOWN), repeat(peelingInfo), repeat(singleLocusMode))
        else:
            for family in jit_families:
                Peeling.peel(family, Peeling.PEEL_DOWN, peelingInfo, singleLocusMode)

    for index, generation in enumerate(reversed(pedigree.generations)):
        print("Peeling Up, Generation", len(pedigree.generations) - index - 1)
        jit_families = [family.toJit() for family in generation.families]

        if args.maxthreads > 1:
            with concurrent.futures.ThreadPoolExecutor(max_workers=nWorkers) as executor:
                 results = executor.map(Peeling.peel, jit_families, repeat(Peeling.PEEL_UP), repeat(peelingInfo), repeat(singleLocusMode))
        else:
            for family in jit_families:
                Peeling.peel(family, Peeling.PEEL_UP, peelingInfo, singleLocusMode)

        sires = set()
        dams = set()
        for family in generation.families:
            sires.add(family.sire)
            dams.add(family.dam)
        updatePosterior(pedigree, peelingInfo, sires, dams)

# updatePosterior updates the posterior term for a specific set of sires and dams.
# The updateSire and updateDam functions perform the updates for a specific sire 
# and specific dam by including all of the information from all of their families.
# This update is currently not multithreaded. It is also currently not ideal -- 
# right now the posterior term is updated for all of the sires/dams no matter 
# whether or not they have been changed since the last update.

def updatePosterior(pedigree, peelingInfo, sires, dams) :

    # if pedigree.mapSireToFamilies is None or pedigree.mapDamToFamilies is None:
    #     pedigree.setupFamilyMap()

    for sire in sires:
        updateSire(sire, peelingInfo)

    for dam in dams:
        updateDam(dam, peelingInfo)


def updateSire(sire, peelingInfo) :

    famList = [fam.idn for fam in sire.families]
    sire = sire.idn
    peelingInfo.posterior[sire,:,:] = 0
    for famId in famList:
        log_update = np.log(peelingInfo.posteriorSire_new[famId,:,:])
        peelingInfo.posterior[sire,:,:] += log_update
        peelingInfo.posteriorSire_minusFam[famId,:,:] = -log_update

    for famId in famList:
        peelingInfo.posteriorSire_minusFam[famId,:,:] += peelingInfo.posterior[sire,:,:]

    #Rescale values.
    peelingInfo.posterior[sire,:,:] = Peeling.expNorm1D(peelingInfo.posterior[sire,:,:])
    peelingInfo.posterior[sire,:,:] /= np.sum(peelingInfo.posterior[sire,:,:], 0)

    for famId in famList:
        peelingInfo.posteriorSire_minusFam[famId,:,:] = Peeling.expNorm1D(peelingInfo.posteriorSire_minusFam[famId,:,:])
        peelingInfo.posteriorSire_minusFam[famId,:,:]  /= np.sum(peelingInfo.posteriorSire_minusFam[famId,:,:], 0)
        
def updateDam(dam, peelingInfo) :

    famList = [fam.idn for fam in dam.families]
    dam = dam.idn
    peelingInfo.posterior[dam,:,:] = 0
    for famId in famList:
        log_update = np.log(peelingInfo.posteriorDam_new[famId,:,:])
        peelingInfo.posterior[dam,:,:] += log_update
        peelingInfo.posteriorDam_minusFam[famId,:,:] = -log_update

    for famId in famList:
        peelingInfo.posteriorDam_minusFam[famId,:,:] += peelingInfo.posterior[dam,:,:]

    peelingInfo.posterior[dam,:,:] = Peeling.expNorm1D(peelingInfo.posterior[dam,:,:])
    peelingInfo.posterior[dam,:,:] /= np.sum(peelingInfo.posterior[dam,:,:], 0)
    for famId in famList:
        peelingInfo.posteriorDam_minusFam[famId,:,:] = Peeling.expNorm1D(peelingInfo.posteriorDam_minusFam[famId,:,:])
        peelingInfo.posteriorDam_minusFam[famId,:,:]  /= np.sum(peelingInfo.posteriorDam_minusFam[famId,:,:], 0)

def getLociAndDistance(snpMap, segMap):
    nSnp = len(snpMap)
    distance = np.full(nSnp, 0, dtype = np.float32)
    loci = np.full((nSnp, 2), 0, dtype = np.int64)

    # Assume snp map and segMap are sorted.
    segIndex = 0
    for i in range(nSnp) :
        pos = snpMap[i]
        # Move along the segMap until we reach a point where we are either at the end of the map, or where the next seg marker occurs after the position in the genotype file.
        # This assumes sorting pretty heavily. Alternative would be to find the neighboring positions in the seg file for each marker in the genotype file.
        while segIndex < (len(segMap)-1) and segMap[segIndex + 1] < pos : 
            segIndex += 1
        
        # Now that positions are known, choose the neighboring markers and the distance to those markers.
        # First two if statements handle the begining and ends of the chromosome.
        if segIndex == 0 and segMap[segIndex] > pos :
            loci[i, :] = (segIndex, segIndex)
            distance[i] = 0
        elif segIndex == (len(segMap)-1) and segMap[segIndex] <= pos :
            loci[i, :] = (segIndex, segIndex)
            distance[i] = 0
        else:
            loci[i,:] = (segIndex, segIndex + 1)
            gap = segMap[segIndex+1] - segMap[segIndex] 
            distance[i] = 1.0 - (pos - segMap[segIndex])/gap #At distance 0, only use segIndex. At distance 1, use segIndex + 1.
    return (loci, distance)


def generateSingleLocusSegregation(peelingInfo, pedigree, args):
    segInfo = None
    if args.segfile is not None:
        # This just gets the locations in the map files.
        snpMap = np.array(InputOutput.readMapFile(args.mapfile, args.startsnp, args.stopsnp)[2])
        segMap = np.array(InputOutput.readMapFile(args.segmapfile)[2])

        loci, distance = getLociAndDistance(snpMap, segMap)
        start = np.min(loci)
        stop = np.max(loci) 
        
        seg = InputOutput.readInSeg(pedigree, args.segfile, start = start, stop = stop)
        loci -= start # Re-align to seg file.
        for i in range(len(distance)):
            segLoc0 = loci[i,0]
            segLoc1 = loci[i,1]
            peelingInfo.segregation[:,:,i] = distance[i]*seg[:,:,segLoc0] + (1-distance[i])*seg[:,:,segLoc1]

    else:
        peelingInfo.segregation[:,:,:] = .25


### ACTUAL PROGRAM BELOW


def getArgs() :
    parser = argparse.ArgumentParser(description='')
    core_parser = parser.add_argument_group("Core arguments")
    core_parser.add_argument('-out', required=True, type=str, help='The output file prefix.')
   
    core_peeling_parser = parser.add_argument_group("Mandatory peeling arguments")
    core_peeling_parser.add_argument('-runtype', default=None, required=False, type=str, help='Program run type. Either "single" or "multi".')

    # Input options
    input_parser = parser.add_argument_group("Input Options")
    InputOutput.add_arguments_from_dictionary(input_parser, InputOutput.get_input_options(), options = ["bfile", "genotypes", "phasefile", "seqfile", "pedigree", "startsnp", "stopsnp"]) 

    # Output options
    output_parser = parser.add_argument_group("Output Options")

    output_parser.add_argument('-no_dosages', action='store_true', required=False, help='Flag to suppress the dosage files.')
    output_parser.add_argument('-no_seg', action='store_true', required=False, help='Flag to suppress the segregation files (e.g. when running for chip imputation and not hybrid peeling).')
    output_parser.add_argument('-no_params', action='store_true', required=False, help='Flag to suppress writing the parameter files.')

    output_parser.add_argument('-haps', action='store_true', required=False, help='Flag to enable writing out the genotype probabilities.')
    output_parser.add_argument('-calling_threshold', default=None, required=False, type=float, nargs="*", help='Genotype calling threshold(s). Multiple space separated values allowed. Use. .3 for best guess genotype.')
    output_parser.add_argument('-binary_call_files', action='store_true', required=False, help='Flag to write out the called genotype files as a binary plink output [Not yet implemented].')
    output_parser.add_argument('-call_phase', action='store_true', required=False, help='Flag to call the phase as well as the genotypes.')
    
    InputOutput.add_arguments_from_dictionary(output_parser, InputOutput.get_output_options(), options = ["writekey", "onlykeyed"]) 


    # Multithreading
    multithread_parser = parser.add_argument_group("Multithreading Options")
    InputOutput.add_arguments_from_dictionary(multithread_parser, InputOutput.get_multithread_options(), options = ["iothreads", "maxthreads"]) 


    peeling_parser = parser.add_argument_group("Optional peeling arguments")
    peeling_parser.add_argument('-ncycles',default=5, required=False, type=int, help='Number of peeling cycles. Default: 5.')
    peeling_parser.add_argument('-length', default=1.0, required=False, type=float, help='Estimated length of the chromosome in Morgans. [Default 1.00]')
    peeling_parser.add_argument('-penetrance',   default=None, required=False, type=str, nargs="*", help=argparse.SUPPRESS) #help='An optional external penetrance file. This will overwrite the default penetrance values.')
    InputOutput.add_arguments_from_dictionary(peeling_parser, InputOutput.get_probability_options(), options = ["error", "seqerror"]) 


    peeling_control_parser = parser.add_argument_group("Peeling control arguments")
    peeling_control_parser.add_argument('-esterrors', action='store_true', required=False, help='Flag to re-estimate the genotyping error rates after each peeling cycle.')
    peeling_control_parser.add_argument('-estmaf', action='store_true', required=False, help='Flag to re-estimate the minor allele frequency after each peeling cycle.')
    peeling_control_parser.add_argument('-nophasefounders', action='store_true', required=False, help='A flag phase a heterozygous allele in one of the founders (if such an allele can be found).')
    peeling_control_parser.add_argument('-sexchrom', action='store_true', required=False, help='A flag to that this is a sex chromosome. Sex needs to be given in the pedigree file. This is currently an experimental option.')


    singleLocus_parser = parser.add_argument_group("Hybrid peeling arguments")
    singleLocus_parser.add_argument('-mapfile',default=None, required=False, type=str, help='a map file for genotype data.')
    singleLocus_parser.add_argument('-segmapfile',default=None, required=False, type=str, help='a map file for the segregation estimates for hybrid peeling.')
    singleLocus_parser.add_argument('-segfile',default=None, required=False, type=str, help='A segregation file for hybrid peeling.')
    # singleLocus_parser.add_argument('-blocksize',default=100, required=False, type=int, help='The number of markers to impute at once. This changes the memory requirements of the program.')




    return InputOutput.parseArgs("AlphaPeel", parser)





def main() :
    args = getArgs()
    pedigree = Pedigree.Pedigree() 
    InputOutput.readInPedigreeFromInputs(pedigree, args)

    singleLocusMode = args.runtype == "single"
    if args.runtype == "multi" and args.segfile :
        print("Running in multi-locus mode, external segfile ignored")

    peelingInfo = PeelingInfo.createPeelingInfo(pedigree, args, phaseFounder = (not args.nophasefounders))

    if singleLocusMode:
        print("Generating seg estimates")
        generateSingleLocusSegregation(peelingInfo, pedigree, args)
    runPeelingCycles(pedigree, peelingInfo, args, singleLocusMode = singleLocusMode)

    PeelingIO.writeGenotypes(pedigree, genoProbFunc = peelingInfo.getGenoProbs)
    if not args.no_params: PeelingIO.writeOutParamaters(peelingInfo)
    if not singleLocusMode and not args.no_seg: InputOutput.writeIdnIndexedMatrix(pedigree, peelingInfo.segregation, args.out + ".seg")


if __name__ == "__main__":
    main()



import datetime
def time_func(text):
    # This creates a decorator with "text" set to "text"
    def timer_dec(func):
        # This is the returned, modified, function
        def timer(*args, **kwargs):
            start_time = datetime.datetime.now()
            values = func(*args, **kwargs)
            print(f"{text}: {(datetime.datetime.now() - start_time).total_seconds():.2f} seconds")
            return values
        return timer

    return timer_dec



import concurrent.futures
from numba import jit, float32, int8, int64, optional, boolean
from numba.experimental import jitclass
import numpy as np
from collections import OrderedDict

from ..tinyhouse import InputOutput
from ..tinyhouse import ProbMath
from ..tinyhouse import HaplotypeOperations

import math

# Defining variables for peel up and peel down. Ideally these would be characters, but numba does not support characters.
PEEL_UP = 0
PEEL_DOWN = 1

# This is the main peeling function.
@jit(nopython=True, nogil=True, locals={'e': float32, 'e4':float32, 'e16':float32, 'e1e':float32})
def peel(family, operation, peelingInfo, singleLocusMode) :

    isSexChrom = peelingInfo.isSexChrom

    e = .000001
    e1e = 1-e
    e4 = e/4
    e16 = e/16

    ### Setup local variables from the peeling information container.
    anterior = peelingInfo.anterior
    penetrance = peelingInfo.penetrance
    posterior = peelingInfo.posterior
    segregation = peelingInfo.segregation
    
    pointSeg = peelingInfo.pointSeg
    segregationTensor = peelingInfo.segregationTensor
    segregationTensor_norm = peelingInfo.segregationTensor_norm

    nLoci = peelingInfo.nLoci
    nOffspring = len(family.offspring)
    sire = family.sire
    dam = family.dam
    fam = family.idn

    #Creating variables here:
    # childToParents: The projection of each child onto the parent genotypes.
    # childSegs: The segregation estimates for a particular child (These are re-used? so need to be stored)
    # allToParents: The projection of each child onto the parental genotypes.
    # parentsMinustChild: The estimate of the parent's genotypes minus the contribution from a specific individual.

    childToParents = np.full((nOffspring, 4, 4, nLoci), 0, dtype = np.float32)
    childSegTensor = np.full((nOffspring, 4, 4, 4, nLoci), 0, dtype = np.float32)
    allToParents = np.full((4, 4, nLoci), 1.0, dtype = np.float32) 
    parentsMinusChild = np.full((nOffspring, 4, 4, nLoci), 1, dtype = np.float32)


    # Some local variables. currentSeg is the segregation estimate of a child (but may be modified).
    currentSeg = np.full((4, nLoci), 1, dtype = np.float32)

    #Construct the joint parent genotypes based on the parent's anterior, penetrance, and posterior terms minus this family.
    
    probSire = anterior[sire,:,:]*penetrance[sire,:,:] * peelingInfo.posteriorSire_minusFam[fam,:,:]
    probDam = anterior[dam,:,:]*penetrance[dam,:,:] * peelingInfo.posteriorDam_minusFam[fam,:,:]

    probSire = probSire/np.sum(probSire, 0)
    probDam = probDam/np.sum(probDam, 0)

    # Einstien sum notation 1: create the joint parental genotypes based on the probabilities for each parent.
    # jointParents = np.einsum("ai, bi -> abi", probSire, probDam)

    jointParents = getJointParents(probSire, probDam)
    jointParents = jointParents/np.sum(np.sum(jointParents, axis = 0), axis = 0)
    jointParents = (1-e)*jointParents + e/16 # There are 4x4 values for each locus in jointparents.

    # There are 4 values for each locus. Normalization is done here so that jointParents is as accurate as possible.
    # We need the posterior terms here for the peeling up step later.
    probSire = probSire*e1e + e4 
    probDam = probDam*e1e + e4 

    # Now construct the parental genotypes based on within-family information. 

    for index in range(nOffspring):
        child = family.offspring[index]
        
        # The child's estimate is the combination of the posterior term and penetrance term for that child.
        # We are estimating the parent's genotypes so the anterior term is ignored to avoid double counting.
        childValues = posterior[child,:,:] * penetrance[child,:,:]
        childValues = childValues/np.sum(childValues, axis = 0)
        childValues = e1e*childValues + e4

        # METHOD 1: Just use the current segregation of the child. 
        currentSeg[:,:] = segregation[child,:,:]
        currentSeg /= np.sum(currentSeg, 0)

        # METHOD 2: Use the segregation estimate of the child minus the contribution at a particular locus.
        # Currently do not recommend using this.
        # if not singleLocusMode :
        #     currentSeg[:,:] = segregation[child,:,:] / pointSeg[child,:,:]
        #     currentSeg /= np.sum(currentSeg, 0)
        # else:
        #     currentSeg[:,:] = segregation[child,:,:]
        
        if isSexChrom and peelingInfo.sex[child] == 0: #0 for male, 1 for female.
            segregationTensor = peelingInfo.segregationTensorXY
        if isSexChrom and peelingInfo.sex[child] == 1: #0 for male, 1 for female.
            segregationTensor = peelingInfo.segregationTensorXX

        #Einstien sum notation 2: Create the child-specific segregation tensor using the child's currrent segregation estimate.
        # childSegTensor[index,:,:,:,:] = np.einsum("abcd, di -> abci", segregationTensor, currentSeg)
        createChildSegs(segregationTensor, currentSeg, childSegTensor[index,:,:,:,:])
    
        #Einstien sum notation 3: Estimate the parental genotypes based on the child's genotypes and their segregation tensor.
        # childToParents[index,:,:,:] = np.einsum("abci, ci -> abi", childSegTensor[index,:,:,:,:], childValues)
        projectChildGenotypes(childSegTensor[index,:,:,:,:], childValues, childToParents[index,:,:,:])
        
    
    # Method 1: estimate the parents genotype and the child-specific posterior terms using iterative normalizing.
    # for i in range(nOffspring) :
    #     parentsMinusChild[i,:,:,:] = jointParents[:,:,:]

    # for i in range(nOffspring):
    #     allToParents *= childToParents[i,:,:,:]
    #     allToParents /= np.sum(np.sum(allToParents, axis = 0), axis=0)

    #     for j in range(nOffspring) :
    #         if i != j :
    #             parentsMinusChild[j,:,:,:] *= childToParents[i,:,:,:]
    #             parentsMinusChild[j,:,:,:] /= np.sum(np.sum(parentsMinusChild[i,:,:,:], axis = 0), axis=0)

    ##
    # Method 2: estimate the parents genotype and the child-specific posterior terms using a log scale.
    ##
    # for i in range(nOffspring) :
    #     parentsMinusChild[i,:,:,:] = np.log(jointParents[:,:,:])
    # allToParents[:,:,:] = 0
    # # #taking out post estimates.
    # for i in range(nOffspring):
    #     log_childToParents = np.log(childToParents[i,:,:,:])
    #     allToParents += log_childToParents
    #     for j in range(nOffspring) :
    #         if i != j :
    #             parentsMinusChild[j,:,:,:] += log_childToParents

    # Method 3: estimate the parents genotype and the child-specific posterior terms using a slightly smarter log scale.

    for i in range(nOffspring) :
        parentsMinusChild[i,:,:,:] = np.log(jointParents[:,:,:])
    allToParents[:,:,:] = 0
    for i in range(nOffspring):
        log_childToParents = np.log(childToParents[i,:,:,:])
        allToParents += log_childToParents
        parentsMinusChild[i,:,:,:] -= log_childToParents # This is done to take away the setimate for an individual child from their parent's posterior term.
    for i in range(nOffspring):
        parentsMinusChild[i,:,:,:] += allToParents

    # Move from a log-scale to a non-log scale and re-normalize.
    allToParents = expNorm2D(allToParents)
    for i in range(nOffspring):
        parentsMinusChild[i,:,:,:] = expNorm2D(parentsMinusChild[i,:,:,:])


    if operation == PEEL_DOWN:
        for i in range(nOffspring):
            child = family.offspring[i]

            #Einstien sum notation 4: Project the parent genotypes down onto the child genotypes.
            # anterior[child,:,:] = np.einsum("abci, abi -> ci", childSegTensor[i,:,:,:,:], parentsMinusChild[i,:,:,:])
            projectParentGenotypes(childSegTensor[i,:,:,:,:], parentsMinusChild[i,:,:,:], anterior[child,:,:])
            anterior[child,:,:] /= np.sum(anterior[child,:,:], 0)
    
    if operation == PEEL_UP :
        # Take the allToParents estimate and combine to estimate the sire and dam's posterior estimates (for this family)

        sirePosterior = combineAndReduceAxis1(allToParents, probDam)
        sirePosterior /= np.sum(sirePosterior, axis = 0)
        sirePosterior = sirePosterior*e1e + e4
        peelingInfo.posteriorSire_new[fam,:,:] = sirePosterior

        damPosterior = combineAndReduceAxis0(allToParents, probSire)
        damPosterior /= np.sum(damPosterior, axis = 0)
        damPosterior = damPosterior*e1e + e4
        peelingInfo.posteriorDam_new[fam,:,:] = damPosterior

    if (not singleLocusMode) and (operation == PEEL_DOWN):
        # Estimate the segregation probabilities for each child.

        for i in range(nOffspring):
            # Child values is the same as in the posterior estimation step above.
            child = family.offspring[i]
            childValues = posterior[child,:,:] * penetrance[child,:,:]
            childValues = childValues/np.sum(childValues, axis = 0)
            childValues = e1e*childValues + e4
            
            if isSexChrom and peelingInfo.sex[child] == 0: #0 for male, 1 for female.
                segregationTensor = peelingInfo.segregationTensorXY
            if isSexChrom and peelingInfo.sex[child] == 1: #0 for male, 1 for female.
                segregationTensor = peelingInfo.segregationTensorXX

            #Einstien sum notation 5:
            # pointSeg[child,:,:] = np.einsum("abcd, abi, ci-> di", segregationTensor, parentsMinusChild[i,:,:,:], childValues)
            #Option 1: Estimate without normalizing.
            # estimateSegregation(segregationTensor, parentsMinusChild[i,:,:,:], childValues, pointSeg[child,:,:])
            #Option 2: Estimate with normalizing. I think this is what we want.
            estimateSegregationWithNorm(segregationTensor, segregationTensor_norm, parentsMinusChild[i,:,:,:], childValues, pointSeg[child,:,:])

            segregation[child,:,:] = (1-e)*collapsePointSeg(pointSeg[child,:,:], peelingInfo.transmissionRate) + e/4

#####
##### The following are a large number of "helper" jit functions that replace the einstien sums in the original scripts.
#####


@jit(nopython=True, nogil=True)
def getJointParents(probSire, probDam):
    # jointParents = np.einsum("ai, bi -> abi", probSire, probDam)
    nLoci = probSire.shape[1]
    output = np.full(shape = (4, 4, nLoci), fill_value = 0, dtype = np.float32)
    for a in range(4) :
        for b in range(4) :
            for i in range(nLoci):
                output[a, b, i] = probSire[a,i] * probDam[b,i]
    return output


@jit(nopython=True, nogil=True)
def createChildSegs(segregationTensor, currentSeg, output):
    # childSegs[index,:,:,:,:] = np.einsum("abcd, di -> abci", segregationTensor, currentSeg)
    nLoci = currentSeg.shape[1]
    output[:,:,:,:] = 0
    for a in range(4) :
        for b in range(4) :
            for c in range(4) :
                for d in range(4) :
                    for i in range(nLoci):
                        output[a, b, c, i] += segregationTensor[a, b, c, d]*currentSeg[d,i]
    
    return output


@jit(nopython=True, nogil=True)
def projectChildGenotypes(childSegs, childValues, output):
    # childToParents[index,:,:,:] = np.einsum("abci, ci -> abi", childSegs[index,:,:,:,:], childValues)
    nLoci = childSegs.shape[3]
    output[:,:,:] = 0
    for a in range(4) :
        for b in range(4) :
            for c in range(4) :
                for i in range(nLoci):
                    output[a, b, i] += childSegs[a, b, c, i]*childValues[c,i]
    
    return output

@jit(nopython=True, nogil=True)
def projectParentGenotypes(childSegs, parentValues, output):
    # anterior[child,:,:] = np.einsum("abci, abi -> ci", childSegs[i,:,:,:,:], parentsMinusChild[i,:,:,:])
    nLoci = childSegs.shape[3]
    output[:,:]=0
    for a in range(4) :
        for b in range(4) :
            for c in range(4) :
                for i in range(nLoci):
                    output[c, i] += childSegs[a,b,c,i]*parentValues[a,b,i]
    
    return output

@jit(nopython=True, nogil=True)
def estimateSegregation(segregationTensor, parentValues, childValues, output):
    # pointSeg[child,:,:] = np.einsum("abcd, abi, ci-> di", segregationTensor, parentsMinusChild[i,:,:,:], childValues)
    nLoci = childValues.shape[1]
    output[:,:]=0
    for a in range(4) :
        for b in range(4) :
            for c in range(4) :
                for d in range(4) :
                    for i in range(nLoci):
                        output[d, i] += segregationTensor[a,b,c,d]*parentValues[a,b,i]*childValues[c,i]
    return output

@jit(nopython=True, nogil=True)
def estimateSegregationWithNorm(segregationTensor, segregationTensor_norm, parentValues, childValues, output):
    # pointSeg[child,:,:] = np.einsum("abcd, abi, ci-> di", segregationTensor, parentsMinusChild[i,:,:,:], childValues)
    nLoci = childValues.shape[1]
    output[:,:]=0
    for a in range(4) :
        for b in range(4) :
            for c in range(4) :
                for d in range(4) :
                    for i in range(nLoci):
                        #Check if norm is 0. Otherwise use norm to normalize.
                        if segregationTensor_norm[a, b, c] != 0:
                            output[d, i] += segregationTensor[a,b,c,d]*parentValues[a,b,i]*childValues[c,i]/segregationTensor_norm[a, b, c]
    return output

@jit(nopython=True, nogil=True)
def combineAndReduceAxis1(jointEstimate, parentEstimate):
    # output = np.einsum("abi, bi-> ai", jointEstimate, parentEstimate)
    nLoci = parentEstimate.shape[1]
    output = np.full((4, nLoci), 0, dtype = np.float32)
    for a in range(4):
        for b in range(4) :
            for i in range(nLoci):
                output[a, i] += jointEstimate[a,b,i]*parentEstimate[b, i]
    return output

@jit(nopython=True, nogil=True)
def combineAndReduceAxis0(jointEstimate, parentEstimate):
    # output = np.einsum("abi, ai-> bi", jointEstimate, parentEstimate)
    nLoci = parentEstimate.shape[1]
    output = np.full((4, nLoci), 0, dtype = np.float32)
    for a in range(4):
        for b in range(4) :
            for i in range(nLoci):
                output[b, i] += jointEstimate[a,b,i]*parentEstimate[a, i]
    return output

@jit(nopython=True, nogil=True)
def expNorm2D(mat):
    # Matrix is 4x4xnLoci: Output is to take the exponential of the matrix and normalize each locus. We need to make sure that there are not any overflow values.
    nLoci = mat.shape[2]
    for i in range(nLoci):
        maxVal = 1 # Log of anything between 0-1 will be less than 0. Using 1 as a default.
        for a in range(4):
            for b in range(4):
                if mat[a, b, i] > maxVal or maxVal == 1:
                    maxVal = mat[a, b, i]
        for a in range(4):
            for b in range(4):
                mat[a, b, i] -= maxVal
    # Normalize.
    tmp = np.exp(mat)
    for i in range(nLoci):
        total = 0
        for a in range(4):
            for b in range(4):
                total += tmp[a, b, i]
        for a in range(4):
            for b in range(4):
                tmp[a, b, i] /= total
    return tmp



@jit(nopython=True, nogil=True)
def expNorm1D(mat):
    # Matrix is 4x4xnLoci: Output is to take the exponential of the matrix and normalize each locus. We need to make sure that there are not any overflow values.
    nLoci = mat.shape[1]
    for i in range(nLoci):
        maxVal = 1 # Log of anything between 0-1 will be less than 0. Using 1 as a default.
        for a in range(4):
            if mat[a, i] > maxVal or maxVal == 1:
                maxVal = mat[a, i]
        for a in range(4):
            mat[a, i] -= maxVal
    tmp = np.exp(mat)
    for i in range(nLoci):
        total = 0
        for a in range(4):
            total += tmp[a,i]
        for a in range(4):
            tmp[a, i] /= total
    return tmp


@jit(nopython=True, nogil=True, locals={'e': float32, 'e2':float32, 'e1e':float32, 'e2i':float32})
def collapsePointSeg(pointSeg, transmission):

    # This is the forward backward algorithm.
    # Segregation estimate state ordering: pp, pm, mp, mm
    nLoci = pointSeg.shape[1] 

    seg = np.full(pointSeg.shape, .25, dtype = np.float32)
    for i in range(nLoci):
        for j in range(4):
            seg[j,i] = pointSeg[j,i]

    tmp = np.full((4), 0, dtype = np.float32)
    new = np.full((4), 0, dtype = np.float32)

    prev = np.full((4), .25, dtype = np.float32)
    for i in range(1, nLoci):
        e = transmission[i-1]
        e2 = e**2
        e1e = e*(1-e)
        e2i = (1.0-e)**2
        for j in range(4):
            tmp[j] = prev[j]*pointSeg[j,i-1]
        
        sum_j = 0
        for j in range(4):
            sum_j += tmp[j]
        for j in range(4):
            tmp[j] = tmp[j]/sum_j

        # !                  fm  fm  fm  fm 
        # !segregationOrder: pp, pm, mp, mm

        new[0] = e2*tmp[3] + e1e*(tmp[1] + tmp[2]) + e2i*tmp[0] 
        new[1] = e2*tmp[2] + e1e*(tmp[0] + tmp[3]) + e2i*tmp[1] 
        new[2] = e2*tmp[1] + e1e*(tmp[0] + tmp[3]) + e2i*tmp[2] 
        new[3] = e2*tmp[0] + e1e*(tmp[1] + tmp[2]) + e2i*tmp[3] 

        # tmp = tmp/np.sum(tmp)
        # new = e2i*tmp + e2 + e1e*(tmp[0] + tmp[3])*same + e1e*(tmp[1] + tmp[2])*diff       

        for j in range(4):
            seg[j,i] *= new[j]
        # seg[:,i] *= new
        prev = new

    prev = np.full((4), .25, dtype = np.float32)
    for i in range(nLoci-2, -1, -1): #zero indexed then minus one since we skip the boundary.
        e = transmission[i]
        e2 = e**2
        e1e = e*(1-e)
        e2i = (1.0-e)**2
        
        for j in range(4):
            tmp[j] = prev[j]*pointSeg[j,i+1]
        
        sum_j = 0
        for j in range(4):
            sum_j += tmp[j]
        for j in range(4):
            tmp[j] = tmp[j]/sum_j

        new[0] = e2*tmp[3] + e1e*(tmp[1] + tmp[2]) + e2i*tmp[0] 
        new[1] = e2*tmp[2] + e1e*(tmp[0] + tmp[3]) + e2i*tmp[1] 
        new[2] = e2*tmp[1] + e1e*(tmp[0] + tmp[3]) + e2i*tmp[2] 
        new[3] = e2*tmp[0] + e1e*(tmp[1] + tmp[2]) + e2i*tmp[3] 

        for j in range(4):
            seg[j,i] *= new[j]
        prev = new
    
    for i in range(nLoci):
        sum_j = 0
        for j in range(4):
            sum_j += seg[j, i]
        for j in range(4):
            seg[j, i] = seg[j, i]/sum_j

    # seg = seg/np.sum(seg, 0)
    return(seg)



from numba import jit
import numpy as np
from . import ProbMath
from . import NumbaUtils
from . HaplotypeLibrary import haplotype_from_indices


def diploidHMM(individual, paternal_haplotypes, maternal_haplotypes, error, recombination_rate, calling_method='dosages', use_called_haps=True, include_geno_probs=False):

    n_loci = len(individual.genotypes)

    # !!!! NEED TO MAKE SURE SOURCE HAPLOTYPES ARE ALL NON MISSING!!!
    if type(paternal_haplotypes) is list or type(paternal_haplotypes) is tuple:
        paternal_haplotypes = np.array(paternal_haplotypes)

    if type(maternal_haplotypes) is list or type(maternal_haplotypes) is tuple:
        maternal_haplotypes = np.array(maternal_haplotypes)

    # Expand error and recombinationRate to arrays as may need to have
    # marker specific error/recombination rates.
    if type(error) is float:
        error = np.full(n_loci, error, dtype=np.float32)
    if type(recombination_rate) is float:
        recombination_rate = np.full(n_loci, recombination_rate, dtype=np.float32)

    # Construct penetrance values (point estimates)
    if use_called_haps:
        point_estimates = getDiploidPointEstimates(individual.genotypes, individual.haplotypes[0], individual.haplotypes[1], paternal_haplotypes, maternal_haplotypes, error)
    elif calling_method in ('sample', 'dosages', 'Viterbi'):
        n_pat = len(paternal_haplotypes)
        n_mat = len(maternal_haplotypes)
        point_estimates = np.ones((n_loci, n_pat, n_mat), dtype=np.float32)
        getDiploidPointEstimates_geno(individual.genotypes, paternal_haplotypes, maternal_haplotypes, error, point_estimates)
    else:
        probs = ProbMath.getGenotypeProbabilities_ind(individual)
        point_estimates = getDiploidPointEstimates_probs(probs, paternal_haplotypes, maternal_haplotypes, error)

    # Do 'sample' and 'Viterbi' before other 'calling_method' as we don't need the forward-backward probs
    if calling_method == 'sample':
        haplotypes = getDiploidSample(point_estimates, recombination_rate, paternal_haplotypes, maternal_haplotypes)
        individual.imputed_haplotypes = haplotypes
        return
    if calling_method == 'Viterbi':
        haplotypes = get_viterbi(point_estimates, recombination_rate, paternal_haplotypes, maternal_haplotypes)
        individual.imputed_haplotypes = haplotypes
        return
    # Run forward-backward algorithm on penetrance values
    total_probs = diploidForwardBackward(point_estimates, recombination_rate)

    if calling_method == 'dosages':
        dosages = getDiploidDosages(total_probs, paternal_haplotypes, maternal_haplotypes)
        individual.dosages = dosages
    if calling_method == 'probabilities':
        values = getDiploidProbabilities(total_probs, paternal_haplotypes, maternal_haplotypes)
        individual.info = values
    if calling_method == 'callhaps':
        raise ValueError('callhaps not yet implimented.')


@jit(nopython=True)
def addDiploidPrior(pointEst, prior):
    nPat, nMat, nLoci = pointEst.shape
    for i in range(nLoci):
        for j in range(nPat):
            for k in range(nMat):
                pointEst[j, k, i] *= prior[j, k]


@jit(nopython=True, nogil=True)
def getDiploidDosages(hapEst, paternalHaplotypes, maternalHaplotypes):
    nPat, nLoci = paternalHaplotypes.shape
    nMat, nLoci = maternalHaplotypes.shape
    dosages = np.full(nLoci, 0, dtype=np.float32)
    for i in range(nLoci):
        for j in range(nPat):
            for k in range(nMat):
                dosages[i] += hapEst[i, j, k]*(paternalHaplotypes[j, i] + maternalHaplotypes[k, i])
    return dosages


@jit(nopython=True)
def getDiploidProbabilities(hapEst, paternalHaplotypes, maternalHaplotypes):
    nPat, nLoci = paternalHaplotypes.shape
    nMat, nLoci = maternalHaplotypes.shape
    probs = np.full((4, nLoci), 0, dtype = np.float32)
    for i in range(nLoci):
        for j in range(nPat):
            for k in range(nMat):
                if paternalHaplotypes[j, i] == 0 and maternalHaplotypes[k, i] == 0:
                    probs[0, i] += hapEst[j, k, i]

                if paternalHaplotypes[j, i] == 0 and maternalHaplotypes[k, i] == 1:
                    probs[1, i] += hapEst[j, k, i]

                if paternalHaplotypes[j, i] == 1 and maternalHaplotypes[k, i] == 0:
                    probs[2, i] += hapEst[j, k, i]

                if paternalHaplotypes[j, i] == 1 and maternalHaplotypes[k, i] == 1:
                    probs[3, i] += hapEst[j, k, i]
    return probs


@jit(nopython=True, nogil=True)
def getDiploidSample(point_estimate, recombination_rate, paternal_haps, maternal_haps):
    """Sample a pair of haplotypes"""
    forward_probs = diploid_forward(point_estimate, recombination_rate, in_place=True)
    haplotypes = diploidSampleHaplotypes(forward_probs, recombination_rate, paternal_haps, maternal_haps)
    return haplotypes


@jit(nopython=True, nogil=True)
def get_viterbi(point_estimates, recombination_rate, paternal_haplotypes, maternal_haplotypes):
    """Get most likely haplotype pair using the Viterbi algorithm"""
    n_loci = point_estimates.shape[0]
    haplotypes = np.full((2, n_loci), 9, dtype=np.int8)

    forward_probs = diploid_forward(point_estimates, recombination_rate)
    paternal_indices, maternal_indices = diploid_viterbi(forward_probs, recombination_rate)
    haplotypes[0] = haplotype_from_indices(paternal_indices, paternal_haplotypes)
    haplotypes[1] = haplotype_from_indices(maternal_indices, maternal_haplotypes)

    return haplotypes


@jit(nopython=True, nogil=True)
def getDiploidPointEstimates(indGeno, indPatHap, indMatHap, paternalHaplotypes, maternalHaplotypes, error):
    nPat, nLoci = paternalHaplotypes.shape
    nMat, nLoci = maternalHaplotypes.shape

    pointEst = np.full((nLoci, nPat, nMat), 1, dtype=np.float32)
    for i in range(nLoci):
        if indGeno[i] != 9:
            for j in range(nPat):
                for k in range(nMat):
                    # Seperate Phased vs non phased loci
                    if indPatHap[i] != 9 and indMatHap[i] != 9:
                        value = 1
                        if indPatHap[i] == paternalHaplotypes[j, i]:
                            value *= (1-error[i])
                        else:
                            value *= error[i]
                        if indMatHap[i] == maternalHaplotypes[k, i]:
                            value *= (1-error[i])
                        else:
                            value *= error[i]
                        pointEst[i, j, k] = value
                    else:
                        #I think this shouldn't be too horrible.
                        sourceGeno = paternalHaplotypes[j, i] + maternalHaplotypes[k, i]
                        if sourceGeno == indGeno[i]:
                            pointEst[i, j, k] = 1-error[i]*error[i]
                        else:
                            pointEst[i, j, k] = error[i]*error[i]
    return pointEst

@jit(nopython=True, nogil=True)
def getDiploidPointEstimates_geno(indGeno, paternalHaplotypes, maternalHaplotypes, error, pointEst):
    nPat, nLoci = paternalHaplotypes.shape
    nMat, nLoci = maternalHaplotypes.shape

    for i in range(nLoci):
        if indGeno[i] != 9:
            error_2 = error[i]*error[i]
            for j in range(nPat):
                for k in range(nMat):

                    #I think this shouldn't be too horrible.
                    sourceGeno = paternalHaplotypes[j, i] + maternalHaplotypes[k, i]
                    if sourceGeno == indGeno[i]:
                        pointEst[i, j, k] = 1-error_2
                    else:
                        pointEst[i, j, k] = error_2


@jit(nopython=True)
def getDiploidPointEstimates_probs(indProbs, paternalHaplotypes, maternalHaplotypes, error):
    nPat, nLoci = paternalHaplotypes.shape
    nMat, nLoci = maternalHaplotypes.shape

    pointEst = np.full((nPat, nMat, nLoci), 1, dtype = np.float32)
    for i in range(nLoci):
        for j in range(nPat):
            for k in range(nMat):
                # I'm just going to be super explicit here.
                p_aa = indProbs[0, i]
                p_aA = indProbs[1, i]
                p_Aa = indProbs[2, i]
                p_AA = indProbs[3, i]
                e = error[i]
                if paternalHaplotypes[j,i] == 0 and maternalHaplotypes[k, i] == 0:
                    value = p_aa*(1-e)**2 + (p_aA + p_Aa)*e*(1-e) + p_AA*e**2

                if paternalHaplotypes[j,i] == 1 and maternalHaplotypes[k, i] == 0:
                    value = p_Aa*(1-e)**2 + (p_aa + p_AA)*e*(1-e) + p_aA*e**2

                if paternalHaplotypes[j,i] == 0 and maternalHaplotypes[k, i] == 1:
                    value = p_aA*(1-e)**2 + (p_aa + p_AA)*e*(1-e) + p_Aa*e**2

                if paternalHaplotypes[j,i] == 1 and maternalHaplotypes[k, i] == 1:
                    value = p_AA*(1-e)**2  + (p_aA + p_Aa)*e*(1-e) + p_aa*e**2

                pointEst[j,k,i] = value
    return pointEst


@jit(nopython=True)
def diploid_normalize(array):
    """Normalize a 'diploid' probability array in place
    The array should have shape: (# paternal haplotypes, # maternal haplotypes, # loci)
    The function normalizes such that the values at each locus sum to 1

    It's possible the accuracy could be improved by using a compensated summation algorithm, e.g.:
    https://en.wikipedia.org/wiki/Kahan_summation_algorithm
    As it stands, running np.sum(array, axis=(0,1)) on the output of this function with an array
    of shape (200,200,1000) gives values that differ from 1 by ~1e-4
    Note also that Numba doesn't support axis=(0,1) so we can't use that"""

    n_pat, n_mat, n_loci = array.shape
    for i in range(n_loci):
        # Use a float64 accumulator to avoid losing precision
        sum_ = np.float64(0)
        for j in range(n_pat):
            for k in range(n_mat):
                sum_ += array[j, k, i]
        for j in range(n_pat):
            for k in range(n_mat):
                array[j, k, i] = array[j, k, i]/sum_


@jit(nopython=True)
def transmit(previous_estimate, recombination_rate, output, pat, mat):
    """Transforms a probability distribution (over pairs of paternal and maternal haplotypes, at a single locus)
    to a probability distribution at the next locus by accounting for emission probabilities (point_estimates)
    and transition probabilities (recombination_rate)

    This is a core step in the forward and backward algorithms

    point_estimates     probability distribution to be transmitted forward. Assume already normalized.
                        shape: (# paternal haplotypes, # maternal haplotypes)
    recombination_rate  recombination rate at this locus - scalar
    output              newly calculated probability distribution over pairs of haplotypes at *this* locus
                        shape: (# paternal haplotypes, # maternal haplotypes)

    Note: previous and estimate are updated by this function"""

    n_pat, n_mat = previous_estimate.shape

    # Get haplotype specific recombinations
    pat[:] = 0
    mat[:] = 0
    for j in range(n_pat):
        for k in range(n_mat):
            pat[j] += previous_estimate[j, k]
            mat[k] += previous_estimate[j, k]

    e = recombination_rate
    e1e = (1-e)*e
    e2m1 = (1-e)**2

    # Adding modifications to pat and mat to take into account number of haplotypes and recombination rate.
    pat *= e1e/n_pat
    mat *= e1e/n_mat

    e2 = e*e/(n_mat*n_pat)

    # Account for recombinations
    for j in range(n_pat):
        for k in range(n_mat):
            output[j, k] = previous_estimate[j, k]*e2m1 + pat[j] + mat[k] + e2


@jit(nopython=True, nogil=True)
def diploid_forward(point_estimate, recombination_rate, in_place=False):
    """Calculate forward probabilities combined with the point_estimates"""

    n_loci, n_pat, n_mat = point_estimate.shape
    if in_place:
        combined = point_estimate
    else:
        combined = point_estimate.copy()  # copy so that point_estimate is not modified

    prev = np.full((n_pat, n_mat), 0.25, dtype=np.float32)

    # Temporary numba variables.
    forward_i = np.empty((n_pat, n_mat), dtype=np.float32)
    tmp_pat = np.empty(n_pat, dtype=np.float32)
    tmp_mat = np.empty(n_mat, dtype=np.float32)

    # Make sure the first locus is normalized.
    combined[0, :, :] /= np.sum(combined[:, :])
    for i in range(1, n_loci):
        # Update estimates at this locus

        # Take the value at locus i-1 and transmit it forward.
        transmit(combined[i-1, :, :], recombination_rate[i], forward_i, tmp_pat, tmp_mat)

        # Combine the forward estimate at locus i with the point estimate at i.
        # This is safe if in_place = True since we have not updated combined[i,:,:] yet and it will be still equal to point_estimate.
        combined[i, :, :] = point_estimate[i, :, :] * forward_i
        combined[i, :, :] /= np.sum(combined[i, :, :])

    return combined


@jit(nopython=True, nogil=True)
def diploid_backward(point_estimate, recombination_rate):
    """Calculate backward probabilities"""

    n_loci, n_pat, n_mat = point_estimate.shape
    backward = np.ones_like(point_estimate, dtype=np.float32)

    prev = np.full((n_pat, n_mat), 0.25, dtype=np.float32)

    # Temporary numba variables.
    combined_i = np.empty((n_pat, n_mat), dtype=np.float32)
    tmp_pat = np.empty(n_pat, dtype=np.float32)
    tmp_mat = np.empty(n_mat, dtype=np.float32)

    for i in range(n_loci-2, -1, -1):
        # Skip the first loci.
        # Combine the backward estimate at i+1 with the point estimate at i+1 (unlike the forward pass, the backward estimate does not contain the point_estimate).
        combined_i[:, :] = backward[i+1, :, :] * point_estimate[i+1, :, :]
        combined_i[:, :] /= np.sum(combined_i)

        # Transmit the combined value forward. This is the backward estimate.
        transmit(combined_i, recombination_rate[i], backward[i, :, :], tmp_pat, tmp_mat)

    return backward


@jit(nopython=True, nogil=True)
def diploidForwardBackward(point_estimate, recombination_rate):
    """Calculate state probabilities at each loci using the forward-backward algorithm"""

    # We may want to split this out into something else.

    est = diploid_backward(point_estimate, recombination_rate)
    est *= diploid_forward(point_estimate, recombination_rate, in_place=True)

    # Return normalized probabilities
    n_loci = est.shape[0]
    for i in range(n_loci):
        est[i, :, :] /= np.sum(est[i, :, :])

    return est


@jit(nopython=True, nogil=True)
def diploidSampleHaplotypes(forward_probs, recombination_rate, paternal_haplotypes, maternal_haplotypes):
    """Sample a pair of paternal and maternal haplotypes from the forward and backward probability distributions
    and paternal and maternal haplotype libraries.
    Returns:
      haplotypes      Pair of haplotypes as a 2D array of shape (2, n_loci)
    """
    n_loci = forward_probs.shape[0]
    haplotypes = np.full((2, n_loci), 9, dtype=np.int8)
    paternal_indices, maternal_indices = diploidOneSample(forward_probs, recombination_rate)
    haplotypes[0] = haplotype_from_indices(paternal_indices, paternal_haplotypes)
    haplotypes[1] = haplotype_from_indices(maternal_indices, maternal_haplotypes)

    return haplotypes


@jit(nopython=True)
def diploidOneSample(forward_probs, recombination_rate):
    """Sample a pair of paternal and maternal haplotypes from the forward and backward probability distributions
    Returns:
      paternal_indices, maternal_indices - arrays of sampled haplotype indices

    A description of the sampling process would be nice here..."""

    n_loci, n_pat, n_mat = forward_probs.shape

    pvals = np.empty((n_pat, n_mat), dtype=np.float32)  # sampled probability distribution at one locus
    paternal_indices = np.empty(n_loci, dtype=np.int64)
    maternal_indices = np.empty(n_loci, dtype=np.int64)

    # Backwards algorithm
    for i in range(n_loci-1, -1, -1): # zero indexed then minus one since we skip the boundary
        # Sample at this locus
        if i == n_loci-1:
            pvals[:, :] = forward_probs[i, :, :]
        else:
            combine_backward_sampled_value(forward_probs[i, :, :], paternal_indices[i+1], maternal_indices[i+1], recombination_rate[i+1], pvals[:, :])

        j, k = NumbaUtils.multinomial_sample_2d(pvals=pvals)
        paternal_indices[i] = j
        maternal_indices[i] = k

    # Last sample (at the first locus)
    j, k = NumbaUtils.multinomial_sample_2d(pvals=pvals)
    paternal_indices[0] = j
    maternal_indices[0] = k

    return paternal_indices, maternal_indices


@jit(nopython=True, nogil=True)
def diploid_viterbi(forward_probs, recombination_rate):
    """Determine the most likely haplotype pair (paternal and maternal) using the Viterbi algorithm
    Returns:
      paternal_indices, maternal_indices - arrays of sampled haplotype indices"""

    n_loci, n_pat, n_mat = forward_probs.shape

    # pvals is the most ikely probability distribution at one locus
    pvals = np.empty((n_pat, n_mat), dtype=np.float32)
    paternal_indices = np.empty(n_loci, dtype=np.int64)
    maternal_indices = np.empty(n_loci, dtype=np.int64)

    # Backwards algorithm
    for i in range(n_loci-1, -1, -1): # zero indexed then minus one since we skip the boundary
        # Sample at this locus
        if i == n_loci-1:
            pvals[:, :] = forward_probs[i, :, :]
        else:
            combine_backward_sampled_value(forward_probs[i, :, :], paternal_indices[i+1], maternal_indices[i+1], recombination_rate[i+1], pvals[:, :])

        idx = np.argmax(pvals)
        j, k = idx//n_mat, idx%n_mat

        paternal_indices[i] = j
        maternal_indices[i] = k

    # Last sample (at the first locus)
    idx = np.argmax(pvals)
    j, k = idx//n_mat, idx%n_mat

    paternal_indices[0] = j
    maternal_indices[0] = k

    return paternal_indices, maternal_indices


@jit(nopython=True)
def diploidIndices(sampled_probs):
    """Get paternal and maternal indices from sampled probability distribution
    Intended to be used with sampled probabilities as returned from diploidOneSample()"""

    n_loci, n_pat, n_mat = sampled_probs.shape
    paternal_indices = np.empty(n_loci, dtype=np.int64)
    maternal_indices = np.empty(n_loci, dtype=np.int64)

    eps = np.finfo(np.float32).eps
    for i in range(n_loci):
        for j in range(n_pat):
            for k in range(n_mat):
                # If sampled_probs[j, k, i] == 1
                # set indices array to values j and k
                if sampled_probs[i, j, k] > 1-eps:
                    paternal_indices[i] = j
                    maternal_indices[i] = k
                    
    return paternal_indices, maternal_indices


@jit(nopython=True)
def combine_backward_sampled_value(previous_estimate, pat_hap, mat_hap, recombination_rate, output):
    """Includes information from the previous sampled locus into the estimate at the current sampled locus.

    previous_estimate   combination of the forward + point estimate at the locus.
                        shape: (# paternal haplotypes, # maternal haplotypes)
    pat_hap, mat_hap    Sampled paternal and maternal haplotypes -- integer
    recombination_rate  recombination rate at this locus - scalar
    output              newly calculated probability distribution over pairs of haplotypes at *this* locus
                        shape: (# paternal haplotypes, # maternal haplotypes)

    Note: previous and estimate are updated by this function"""

    n_pat, n_mat = previous_estimate.shape

    e = recombination_rate
    single_rec = (1-e)*e
    no_rec = (1-e)**2
    double_rec = e*e

    # Haplotype is moving from pat_hap, mat_hap.
    # Double recombination -- both haplotypes change.
    output[:, :] = double_rec/(n_mat*n_pat)

    # Maternal recombination -- pat_hap stays the same.
    for k in range(n_mat):
        output[pat_hap, k] += single_rec/n_mat

    # Paternal recombination -- mat_hap stays the same.
    for j in range(n_pat):
        output[j, mat_hap] += single_rec/n_pat

    # No recombinations -- both haplotypes stay the same.
    output[pat_hap, mat_hap] += no_rec

    # Add in forward_plus_combined
    output *= previous_estimate
    return output



import numpy as np
import numba
import sys
from numba import jit, int8, int64, boolean, deferred_type, optional, float32, double
from numba.experimental import jitclass

from collections import OrderedDict
from . import InputOutput
from . import ProbMath
from . import MultiThreadIO

class Family(object):
    """Family is a container for fullsib families"""
    def __init__(self, idn, sire, dam, offspring):
        self.idn = idn
        self.sire = sire
        self.dam = dam
        self.offspring = offspring
        self.generation = max(sire.generation, dam.generation) + 1

        # Add this family to both the sire and dam's family.
        self.sire.families.append(self)
        self.dam.families.append(self)

    def addChild(self, child) :
        self.offspring.append(child)

    def toJit(self):
        """Returns a just in time version of itself with Individuals replaced by id numbers"""
        offspring = np.array([child.idn for child in self.offspring])
        return jit_Family(self.idn, self.sire.idn, self.dam.idn, offspring)


spec = OrderedDict()
spec['idn'] = int64
spec['sire'] = int64
spec['dam'] = int64
spec['offspring'] = int64[:]

@jitclass(spec)
class jit_Family(object):
    def __init__(self, idn, sire, dam, offspring):
        self.idn = idn
        self.sire = sire
        self.dam = dam
        self.offspring = offspring.astype(np.int64)

class Individual(object):
    
    def __init__(self, idx, idn) :

        self.genotypes = None
        self.haplotypes = None
        self.dosages = None

        self.imputed_genotypes = None
        self.imputed_haplotypes = None

        self.reads = None
        self.longReads = []

        # Do we use this?
        self.genotypeDosages = None
        self.haplotypeDosages = None
        self.hapOfOrigin = None

        # Info is here to provide other software to add in additional information to an Individual.
        self.info = None

        #For plant impute. Inbred is either DH or heavily selfed. Ancestors is historical source of the cross (may be more than 2 way so can't handle via pedigree).
        self.inbred = False
        self.imputationAncestors = [] #This is a list of lists. Either length 0, length 1 or length 2.
        self.selfingGeneration = None


        self.sire = None
        self.dam = None
        self.idx = idx # User inputed string identifier
        self.idn = idn # ID number assigned by the pedigree
        self.fileIndex = dict() # Position of an animal in each file when reading in. Used to make sure Input and Output order are the same.
        self.fileIndex["id"] = idx

        self.dummy = False

        self.offspring = []
        self.families = []

        self.sex = -1
        self.generation = None

        self.initHD = False
        # used in pythonHMM, but how should this best be coded when already have initHD? Should probably be set when data is read in,
        # but need high_density_threshold to be set in the pedigree first
        self.is_high_density = False

        self.genotypedFounderStatus = None #?
    
    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.idn)

    def subset(self, start, stop):
        # Take an individual and create an individual that just contains information on those markers.
        # It's okay if this wipes other information.

        # print(self.__class__)
        new_ind = self.__class__(self.idx, self.idn)
        if self.genotypes is not None:
            new_ind.genotypes = self.genotypes[start:stop].copy() # Maybe could get away with not doing copies... doing them just to be safe.

        if self.haplotypes is not None:
            new_ind.haplotypes = (self.haplotypes[0][start:stop].copy(), self.haplotypes[1][start:stop].copy())

        if self.reads is not None:
            new_ind.reads = (self.reads[0][start:stop].copy(), self.reads[1][start:stop].copy())
        return new_ind

    def getPercentMissing(self):
        return np.mean(self.genotypes == 9)

    def getGeneration(self):

        if self.generation is not None : return self.generation

        if self.dam is None: 
            damGen = -1
        else:
            damGen = self.dam.getGeneration()
        if self.sire is None: 
            sireGen = -1
        else:
            sireGen = self.sire.getGeneration()

        self.generation = max(sireGen, damGen) + 1
        return self.generation


    def constructInfo(self, nLoci, genotypes = True,  haps = False, reads = False) :
        if genotypes and self.genotypes is None:
            self.genotypes = np.full(nLoci, 9, dtype = np.int8)
        
        if  haps and self.haplotypes is None:
            self.haplotypes = (np.full(nLoci, 9, dtype = np.int8), np.full(nLoci, 9, dtype = np.int8))

        if reads and self.reads is None:
            self.reads = (np.full(nLoci, 0, dtype = np.int64), np.full(nLoci, 0, dtype = np.int64))
    
    def isFounder(self):
        return (self.sire is None) and (self.dam is None)

    def getGenotypedFounderStatus(self):
        # options: 1: "GenotypedFounder", 0:"ChildOfNonGenotyped", 2:"ChildOfGenotyped"
        if self.genotypedFounderStatus is None:
            if self.isFounder() :
                if self.genotypes is None or np.all(self.genotypes == 9):
                    self.genotypedFounderStatus = 0 
                else:
                    self.genotypedFounderStatus = 1
            else:
                parentStatus = max(self.sire.getGenotypedFounderStatus(), self.sire.getGenotypedFounderStatus())
                if parentStatus > 0:
                    self.genotypedFounderStatus = 2
                else:
                    if self.genotypes is None or np.all(self.genotypes == 9):
                        self.genotypedFounderStatus = 0 
                    else:
                        self.genotypedFounderStatus = 1

        return self.genotypedFounderStatus
    def isGenotypedFounder(self):
        return (self.getGenotypedFounderStatus() == 1)


class PlantImputeIndividual(Individual):
    """Simple derived class for AlphaPlantImpute2
    with some extra member variables"""
    def __init__(self, idx, idn):
        super().__init__(idx, idn)
        self.founders = []
        self.descendants = []


# Not sure of the code source: https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/
# Slightly modified.
import re 
def sorted_nicely( l , key): 
    """ Sort the given iterable in the way that humans expect.""" 
    return sorted(l, key = lambda k: alphanum_key(key(k)))

def alphanum_key(k):
    convert = lambda text: int(text) if text.isdigit() else text 
    return [ convert(c) for c in re.split('([0-9]+)', str(k)) ] 


class Generation(object):

    def __init__(self, number):

        self.number = number

        self.families = []

        self.individuals = []
        self.sires = set()
        self.dams = set()
        self.parents = set()

    def add_individual(self, ind):
        self.individuals.append(ind)

    def add_family(self, fam):
        self.families.append(fam)
        self.sires.add(fam.sire)
        self.dams.add(fam.dam)
        self.parents.add(fam.sire)
        self.parents.add(fam.dam)
        # Note: Individuals are added seperately.



class Pedigree(object):
 
    def __init__(self, fileName = None, constructor = Individual):

        self.maxIdn = 0
        self.maxFam = 0

        self.individuals = dict()
        self.families = None
        self.constructor = constructor
        self.nGenerations = 0
        self.generations = None #List of lists

        self.truePed = None
        self.nLoci = 0
        
        self.startsnp = 0
        self.endsnp = self.nLoci

        self.referencePanel = [] #This should be an array of haplotypes. Or a dictionary?

        self.maf=None #Maf is the frequency of 2s.

        # Threshold that determines if an individual is high-density genotyped
        self.high_density_threshold = 0.9

        if fileName is not None:
            self.readInPedigree(fileName)

        self.args = None
        self.writeOrderList = None

        self.allele_coding = None

    def reset_families(self):

        for ind in self.individuals.values():
            ind.families = []
            ind.generation = None

        self.nGenerations = 0
        self.generations = None
        self.families = None
        for ind in self:
            ind.families = []

        self.setupFamilies()

    def subset(self, start, stop):
        new_pedigree = Pedigree(constructor = self.constructor)
        new_pedigree.nLoci = stop - start
        # Add all of the individuals.

        for ind in self:
            # Note: ind.subset strips away all of the family information.
            new_pedigree[ind.idx] = ind.subset(start, stop)

        for ind in self:
            if ind.sire is not None:
                new_ind = new_pedigree[ind.idx]
                new_sire = new_pedigree[ind.sire.idx]


                # Add individuals
                new_ind.sire = new_sire
                new_sire.offspring.append(new_ind)

            if ind.dam is not None:
                new_ind = new_pedigree[ind.idx]
                new_dam = new_pedigree[ind.dam.idx]

                # Add individuals
                new_ind.dam = new_dam
                new_dam.offspring.append(new_ind)
        return new_pedigree

    def merge(self, new_pedigree, start, stop):
        # This just merged genotype, haplotype, and dosage information (if availible).
        
        for ind in self:
            new_ind = new_pedigree[ind.idx]
            if new_ind.genotypes is not None:
                if ind.genotypes is None:
                    ind.genotypes = np.full(self.nLoci, 9, dtype = np.int8)
                ind.genotypes[start:stop] = new_ind.genotypes
            
            if new_ind.dosages is not None:
                if ind.dosages is None:
                    ind.dosages = np.full(self.nLoci, -1, dtype = np.float32)

                ind.dosages[start:stop] = new_ind.dosages
            
            if new_ind.haplotypes is not None:
                if ind.haplotypes is None:
                    ind.haplotypes = (np.full(self.nLoci, 9, dtype = np.int8), np.full(self.nLoci, 9, dtype = np.int8))

                ind.haplotypes[0][start:stop] = new_ind.haplotypes[0]
                ind.haplotypes[1][start:stop] = new_ind.haplotypes[1]



    def __len__(self):
        return len(self.individuals)


    def writeOrder(self):
        if self.writeOrderList is None:
            inds = [ind for ind in self if (not ind.dummy) and (self.args.writekey in ind.fileIndex)]
            self.writeOrderList = sorted_nicely(inds, key = lambda ind: ind.fileIndex[self.args.writekey])
            
            if not self.args.onlykeyed:
                indsNoDummyNoFileIndex = [ind for ind in self if (not ind.dummy) and (not self.args.writekey in ind.fileIndex)]
                self.writeOrderList.extend(sorted_nicely(indsNoDummyNoFileIndex, key = lambda ind: ind.idx))
                
                dummys = [ind for ind in self if ind.dummy]
                self.writeOrderList.extend(sorted_nicely(dummys, key = lambda ind: ind.idx))

        for ind in self.writeOrderList :
            yield (ind.idx, ind)

    def setMaf(self):
        """Calculate minor allele frequencies at each locus"""

        # The default values of 1 (maf) and 2 (counts) provide a sensible prior
        # For example, a locus where all individuals are missing, the MAF will be 0.5
        maf = np.full(self.nLoci, 1, dtype = np.float32)
        counts = np.full(self.nLoci, 2, dtype = np.float32)
        for ind in self.individuals.values():
            if ind.genotypes is not None:
                addIfNotMissing(maf, counts, ind.genotypes)
        
        self.maf = maf/counts

    def getMissingness(self):
        missingness = np.full(self.nLoci, 1, dtype = np.float32)

        counts = 0
        for ind in self.individuals.values():
            if ind.genotypes is not None:
                counts += 1
                addIfMissing(missingness, ind.genotypes)
        return missingness/counts


    def set_high_density(self):
        """Set whether each individual is high-density"""
        for individual in self:#.individuals.values():
            is_high_density = np.mean(individual.genotypes != 9) >= self.high_density_threshold
            if is_high_density:
                individual.is_high_density = True


    def fillIn(self, genotypes = True, haps = False, reads = False):

        for individual in self:
            individual.constructInfo(self.nLoci, genotypes = True, haps = haps, reads = reads)


    def __getitem__(self, key) :
        return self.individuals[key]

    def __setitem__(self, key, value):
        self.individuals[key] = value

    def __iter__(self) :
        if self.generations is None:
            self.setUpGenerations()
        for gen in self.generations:
            for ind in gen.individuals:
                yield ind

    def __reversed__(self) :
        if self.generations is None:
            self.setUpGenerations()
        for gen in reversed(self.generations):
            for ind in gen.individuals:
                yield ind

    def sort_individuals(self, individuals):

        return {k:v for k, v in sorted(individuals.items(), key = lambda pair: alphanum_key(pair[0]))}

    # Generation code

    def setUpGenerations(self) :
        # Try and make pedigree independent.
        self.individuals = self.sort_individuals(self.individuals)

        self.nGenerations = 0
        #We can't use a simple iterator over self here, becuase __iter__ calls this function.
        for idx, ind in self.individuals.items():
            gen = ind.getGeneration()
            self.nGenerations = max(gen, self.nGenerations)
        self.nGenerations += 1 #To account for generation 0. 

        self.generations = [Generation(i) for i in range(self.nGenerations)]

        for idx, ind in self.individuals.items():
            gen = ind.getGeneration()
            self.generations[gen].add_individual(ind)

    #This is really sloppy, but probably not important.
    def setupFamilies(self) :

        if self.generations is None:
            self.setUpGenerations()

        self.families = dict()
        for ind in self:
            if not ind.isFounder():
                parents = (ind.sire.idx, ind.dam.idx)
                if parents in self.families :
                    self.families[parents].addChild(ind)
                else:
                    self.families[parents] = Family(self.maxFam, ind.sire, ind.dam, [ind])
                    self.maxFam += 1

        
        for family in self.families.values():
            self.generations[family.generation].add_family(family)
    

    def getFamilies(self, rev = False) :
        if self.generations is None:
            self.setUpGenerations()
        if self.families is None:
            self.setupFamilies()

        gens = range(0, len(self.generations))
        if rev: gens = reversed(gens)

        for i in gens:
            for family in self.generations[i].families:
                yield family
 


    def getIndividual(self, idx) :
        if idx not in self.individuals:
            self.individuals[idx] = self.constructor(idx, self.maxIdn)
            self.maxIdn += 1
            self.generations = None
        return self.individuals[idx]

    def readInPedigree(self, fileName):
        with open(fileName) as f:
            lines = f.readlines()
        pedList = [line.split() for line in lines]
        self.readInPedigreeFromList(pedList)

    def readInPlantInfo(self, fileName):
        with open(fileName) as f:
            lines = f.readlines()

        for line in lines:
            parts = line.split()
            idx = parts[0]; 

            if idx not in self.individuals:
                self.individuals[idx] = self.constructor(idx, self.maxIdn)
                self.maxIdn += 1

            ind = self.individuals[idx]
            if len(parts) > 1:
                if parts[1] == "DH" or parts[1] == "INBRED":
                    ind.inbred = True
                elif parts[1][0] == "S" :
                    ind.inbred = False
                    ind.selfingGeneration = int(parts[1][1:])
                else:
                    ind.inbred = False

            if len(parts) > 2:
                if "|" in line:
                    first, second = line.split("|")
                    self.addAncestors(ind, first.split()[2:])
                    self.addAncestors(ind, second.split())
                else:
                    self.addAncestors(ind, parts[2:])


    def addAncestors(self, ind, parts):
        ancestors = []
        for idx in parts:
            if idx not in self.individuals:
                self.individuals[idx] = self.constructor(idx, self.maxIdn)
                self.maxIdn += 1
            ancestor = self.individuals[idx]
            ancestors.append(ancestor)
        ind.imputationAncestors.append(ancestors)


    def readInPedigreeFromList(self, pedList):
        index = 0
        for parts in pedList :
            idx = parts[0]
            self.individuals[idx] = self.constructor(idx, self.maxIdn)
            self.maxIdn += 1
            self.individuals[idx].fileIndex['pedigree'] = index; index += 1

        for parts in pedList :
            idx = parts[0]
            if parts[1] == "0": parts[1] = None
            if parts[2] == "0": parts[2] = None
            
            if parts[1] is not None and parts[2] is None:
                parts[2] = "MotherOf"+parts[0]
            if parts[2] is not None and parts[1] is None:
                parts[1] = "FatherOf"+parts[0] 

            ind = self.individuals[parts[0]]
            
            if parts[1] is not None:
                if parts[1] not in self.individuals:
                    self.individuals[parts[1]] = self.constructor(parts[1], self.maxIdn)
                    self.maxIdn += 1
                    self.individuals[parts[1]].fileIndex['pedigree'] = index; index += 1
                    self.individuals[parts[1]].dummy=True

                sire = self.individuals[parts[1]]
                ind.sire = sire
                sire.offspring.append(ind)
                sire.sex = 0

            if parts[2] is not None:
                if parts[2] not in self.individuals:
                    self.individuals[parts[2]] = self.constructor(parts[2], self.maxIdn)
                    self.maxIdn += 1
                    self.individuals[parts[2]].fileIndex['pedigree'] = index; index += 1
                    self.individuals[parts[1]].dummy=True

                dam = self.individuals[parts[2]]
                ind.dam = dam
                dam.offspring.append(ind)
                dam.sex = 1

            # Optional fourth column contains sex OR inbred/outbred status
            if len(parts) > 3:
                male, female = {'m', '0', 'xy'}, {'f', '1', 'xx'}
                inbred, outbred = {'dh', 'inbred'}, {'outbred'}
                expected_entries = male | female | inbred | outbred
                if parts[3].lower() not in expected_entries:
                    print(f"ERROR: unexpected entry in pedigree file, fourth field: '{parts[3]}'\nExiting...")
                    sys.exit(2)
                # Sex
                if parts[3].lower() in male:
                    ind.sex = 0
                elif parts[3].lower() in female:
                    ind.sex = 1
                # Inbred/DH
                if parts[3].lower() in inbred:
                    ind.inbred = True
                elif parts[3].lower() in outbred:
                    ind.inbred = False


    def readInFromPlink(self, idList, pedList, bed, externalPedigree = False):
        index = 0

        if not externalPedigree:
            self.readInPedigreeFromList(pedList)
    
        for i, idx in enumerate(idList):
            genotypes=bed[:, i].copy() ##I think this is the right order. Doing the copy to be safe.
            nLoci = len(genotypes)
            if self.nLoci == 0:
                self.nLoci = nLoci
            if self.nLoci != nLoci:
                print(f"ERROR: incorrect number of loci when reading in plink file. Expected {self.nLoci} got {nLoci}.\nExiting...")
                sys.exit(2)
            if idx not in self.individuals:
                self.individuals[idx] = self.constructor(idx, self.maxIdn)
                self.maxIdn += 1

            ind = self.individuals[idx]
            ind.constructInfo(nLoci, genotypes=True)
            ind.genotypes = genotypes

            ind.fileIndex['plink'] = index; index += 1

            if np.mean(genotypes == 9) < .1 :
                ind.initHD = True


    
    def check_line(self, id_data, fileName, idxExpected=None, ncol=None, getInd=True, even_cols=False):
        idx, data = id_data

        if idxExpected is not None and idx != idxExpected:
            print(f"ERROR: Expected individual {idxExpected} but got individual {idx}.\nExiting...")
            sys.exit(2)
        if ncol is None:
            ncol = len(data)
        if ncol != len(data):
            print(f"ERROR: incorrect number of columns in {fileName}. Expected {ncol} values but got {len(data)} for individual {idx}.\nExiting...")
            sys.exit(2)
        if even_cols and ncol % 2 != 0:
            print(f"ERROR: file {fileName} doesn't contain an even number of allele columns for individual {idx}.\nExiting...")
            sys.exit(2)
        
        nLoci = len(data)
        if self.nLoci == 0:
            self.nLoci = nLoci
        if self.nLoci != nLoci:
            print(f"ERROR: inconsistent number of markers or alleles in {fileName}. Expected {self.nLoci} got {nLoci}.")
            sys.exit(2)

        ind = None
        if getInd :
            ind = self.getIndividual(idx)

        return ind, data, ncol 


    def update_allele_coding(self, alleles):
        """Update allele codings with new alleles
        self.allele_coding      - array of shape (2, n_loci) such that:
        self.allele_coding[0]   - array of alleles that are coded as 0
                                  (these are set to the first alleles 'seen')
        self.allele_coding[1]   - array of alleles that are coded as 1
                                  (these are alleles that are different from
                                  self.allele_coding[0])
        alleles  - alleles as read in from PLINK file
                array of dtype np.bytes_, b'0' is 'missing'

        This function is much like finding unique entries in a list:
        only add a new item if it is different from those seen before
        In this case, only record the first two uniques found,
        but also check there are only 2 alleles in total"""

        # If allele coding is complete then skip the update
        if self.allele_coding_complete():
            return

        # Update any missing entries in self.allele_coding[0]
        mask = self.allele_coding[0] == b'0'
        self.allele_coding[0, mask] = alleles[mask]

        # Update entries in self.allele_coding[1] if:
        # - the alleles have not already been seen in self.allele_coding[0]
        # - and the entry (in self.allele_coding[1]) is missing
        mask = self.allele_coding[1] == b'0'
        mask &= self.allele_coding[0] != alleles
        self.allele_coding[1, mask] = alleles[mask]

        # Check for > 2 alleles at any loci. These are:
        # - alleles that are not missing
        # - alleles that are not in either of self.allele_coding[0] or self.allele_coding[1]
        mask = alleles != b'0'
        mask &= ((alleles != self.allele_coding[0]) & (alleles != self.allele_coding[1]))  # poss speedup alleles != self.allele_coding[0] done above
        if np.sum(mask) > 0:
            print(f'ERROR: more than two alleles found in input file(s) at loci {np.flatnonzero(mask)}\nExiting...')
            sys.exit(2)


    def allele_coding_complete(self):
        """Check whether the allele coding is complete (contains no missing values)"""
        if self.allele_coding is None:
            return False
        else:
            return np.sum(self.allele_coding == b'0') == 0


    def decode_alleles(self, alleles):
        """Decode PLINK plain text alleles to AlphaGenes genotypes or haplotypes
        handles single individuals - alleles has shape (n_loci*2, )
        or multiple individuals    - alleles has shape (n_individuals, n_loci*2)"""
        # 'Double' self.allele_coding as there are two allele columns at each locus
        coding = np.repeat(self.allele_coding, 2, axis=1)

        decoded = np.full_like(alleles, b'0', dtype=np.int8)
        decoded[alleles == coding[0]] = 0  # set alleles coded as 0
        decoded[alleles == coding[1]] = 1  # set alleles coded as 1
        decoded[alleles == b'0'] = 9       # convert missing (b'0' -> 9)

        # Extract haplotypes
        decoded = np.atleast_2d(decoded)
        n_haps = decoded.shape[0] * 2
        n_loci = decoded.shape[1] // 2
        haplotypes = np.full((n_haps, n_loci), 9, dtype=np.int8)
        haplotypes[::2] = decoded[:, ::2]
        haplotypes[1::2] = decoded[:, 1::2]
        
        # Convert to genotypes
        genotypes = decoded[:, ::2] + decoded[:, 1::2]
        genotypes[genotypes > 9] = 9  # reset missing values

        return genotypes.squeeze(), haplotypes


    def encode_alleles(self, haplotypes):
        """Encode haplotypes as PLINK plain text
        handles any even number of haplotypes with shape (n_individuals*2, n_loci)"""
        assert len(haplotypes) % 2 == 0
        assert len(haplotypes[0])== self.nLoci

        # 'Double' self.allele_coding as there are two allele columns at each locus in PLINK format
        coding = np.repeat(self.allele_coding, 2, axis=1)

        # Encoded array is 'reshaped' - one individual per line, each locus is a pair of alleles
        encoded = np.full((len(haplotypes)//2, self.nLoci*2), b'0', dtype=np.bytes_)
        # 'Splice' haplotypes into (adjacent) pairs of alleles
        encoded[:, ::2] = haplotypes[::2]
        encoded[:, 1::2] = haplotypes[1::2]

        # Encode
        mask0 = encoded == b'0'  # major alleles (0)
        mask1 = encoded == b'1'  # minor alleles (1)
        mask9 = encoded == b'9'  # missing (9)
        encoded[mask0] = np.broadcast_to(coding[0], encoded.shape)[mask0]
        encoded[mask1] = np.broadcast_to(coding[1], encoded.shape)[mask1]
        encoded[mask9] = b'0'

        return encoded.squeeze()


    def check_allele_coding(self, filename):
        """Check coding is sensible"""
        # Monoallelic loci: 
        #   allele_coding[0] filled, but allele_coding[1] unfilled, i.e. coding[1] == b'0'
        n_monoallelic = (self.allele_coding[1] == b'0').sum()
        # Unusual letters
        unusual = ~np.isin(self.allele_coding, [b'A', b'C', b'G', b'T', b'0'])  # unexpected letters
        if np.sum(unusual) > 0:
            letters = ' '.join(np.unique(self.allele_coding[unusual].astype(str)))
            print(f'ERROR: unexpected values found in {filename}: [{letters}].\n'
                  f'Please check the file is in PLINK .ped format\nExiting...')
            sys.exit(2)
        elif n_monoallelic > 0:
            print(f'WARNING: allele coding from {filename} has {n_monoallelic} monoallelic loci')
        else:
            # Reassuring message if tests pass
            print('Allele coding OK')


    def readInPed(self, filename, startsnp=None, stopsnp=None, haps=False, update_coding=False):
        """Read in genotypes, and optionally haplotypes, from a PLINK plain text formated file, usually .ped
        If update_coding is True, the allele coding is interpreted from the .ped file and any coding
        in self.allele_coding is updated (if the coding is incomplete).
        Note: to force a full read of the allele coding set self.allele_codine = None first"""

        print(f'Reading in PLINK .ped format: {filename}')
        # Check the allele coding is to be got from file or is provided via self.allele_coding
        if not update_coding and self.allele_coding is None:
            raise ValueError('readInPed () called with no allele coding')

        data_list = MultiThreadIO.readLinesPlinkPlainTxt(filename, startsnp=startsnp, stopsnp=stopsnp, dtype=np.bytes_)
        index = 0
        ncol = None
        if self.nLoci != 0:
            # Temporarilly double nLoci while reading in PLINK plain text formats (two fields per locus)
            # otherwise reading of multiple PLINK files results in an 'Incorrect number of values'
            # error in check_line()
            self.nLoci = self.nLoci * 2

        if not self.allele_coding_complete():
            if self.allele_coding is None:
                print(f'Interpreting allele coding from {filename}')
            else:
                print(f'Updating allele coding with coding from {filename}')

        for value in data_list:
            ind, alleles, ncol = self.check_line(value, filename, idxExpected=None, ncol=ncol, even_cols=True)
            ind.constructInfo(self.nLoci, genotypes=True)
            ind.fileIndex['plink'] = index; index += 1

            if update_coding:
                # Initialise allele coding array if none exists
                # read_or_create = 'Reading' if self.allele_coding is None else 'Updating'
                if self.allele_coding is None:
                    self.allele_coding = np.full((2, self.nLoci//2), b'0', dtype=np.bytes_)

                # Update allele codes
                # print(f'{read_or_create} allele coding with coding from {filename}')
                self.update_allele_coding(alleles[::2])   # first allele in each pair
                self.update_allele_coding(alleles[1::2])  # second allele

            # Decode haplotypes and genotypes
            ind.genotypes, haplotypes = self.decode_alleles(alleles)
            if haps:
                ind.haplotypes = haplotypes

            if np.mean(ind.genotypes == 9) < .1:
                ind.initHD = True

        # Check allele coding
        self.check_allele_coding(filename)

        # Reset nLoci back to its undoubled state
        self.nLoci = self.nLoci//2


    def readInGenotypes(self, fileName, startsnp=None, stopsnp = None):

        print("Reading in AlphaGenes format:", fileName)
        index = 0
        ncol = None

        data_list = MultiThreadIO.readLines(fileName, startsnp = startsnp, stopsnp = stopsnp, dtype = np.int8)

        for value in data_list:
            ind, genotypes, ncol = self.check_line(value, fileName, idxExpected = None, ncol = ncol)

            ind.constructInfo(self.nLoci, genotypes=True)
            ind.genotypes = genotypes

            ind.fileIndex['genotypes'] = index; index += 1

            if np.mean(genotypes == 9) < .1 :
                ind.initHD = True

    def readInReferencePanel(self, fileName, startsnp=None, stopsnp = None):

        print("Reading in reference panel:", fileName)
        index = 0
        ncol = None

        data_list = MultiThreadIO.readLines(fileName, startsnp = startsnp, stopsnp = stopsnp, dtype = np.int8)

        for value in data_list:
            ind, haplotype, ncol = self.check_line(value, fileName, idxExpected = None, ncol = ncol, getInd=False)
            self.referencePanel.append(haplotype)

    def readInPhase(self, fileName, startsnp=None, stopsnp = None):
        index = 0
        ncol = None

        data_list = MultiThreadIO.readLines(fileName, startsnp = startsnp, stopsnp = stopsnp, dtype = np.int8)

        e = 0
        currentInd = None
        for value in data_list:

            if e == 0: 
                idxExpected = None
            else:
                idxExpected = currentInd.idx

            ind, haplotype, ncol = self.check_line(value, fileName, idxExpected = idxExpected, ncol = ncol)
            currentInd = ind

            ind.constructInfo(self.nLoci, haps=True)
            ind.haplotypes[e][:] = haplotype
            e = 1-e

            ind.fileIndex['phase'] = index; index += 1

        
    def readInSequence(self, fileName, startsnp=None, stopsnp = None):
        index = 0
        ncol = None

        print("Reading in sequence data :", fileName)
        
        data_list = MultiThreadIO.readLines(fileName, startsnp = startsnp, stopsnp = stopsnp, dtype = np.int64)
        e = 0
        currentInd = None

        for value in data_list:
            if e == 0: 
                idxExpected = None
            else:
                idxExpected = currentInd.idx

            ind, reads, ncol = self.check_line(value, fileName, idxExpected = idxExpected, ncol = ncol)
            currentInd = ind

            ind.constructInfo(self.nLoci, reads=True)
            ind.fileIndex['sequence'] = index; index += 1

            ind.reads[e][:] = reads
            e = 1-e


    def callGenotypes(self, threshold):
        for idx, ind in self.writeOrder():
            matrix = ProbMath.getGenotypeProbabilities_ind(ind, InputOutput.args)
            
            matrixCollapsedHets = np.array([matrix[0,:], matrix[1,:] + matrix[2,:], matrix[3,:]], dtype=np.float32)
            calledGenotypes = np.argmax(matrixCollapsedHets, axis = 0)
            setMissing(calledGenotypes, matrixCollapsedHets, threshold)
            if InputOutput.args.sexchrom and ind.sex == 0:
                doubleIfNotMissing(calledGenotypes)
            ind.genotypes = calledGenotypes


    def writePedigree(self, outputFile):
        with open(outputFile, 'w+') as f:
            for ind in self:
                sire = "0"
                if ind.sire is not None:
                    sire = ind.sire.idx
                dam = "0"
                if ind.dam is not None:
                    dam = ind.dam.idx
                f.write(ind.idx + ' ' + sire + ' ' + dam + '\n')


    def writeGenotypes(self, outputFile):

        data_list = []
        for ind in self :
            data_list.append( (ind.idx, ind.genotypes) )

        MultiThreadIO.writeLines(outputFile, data_list, str)


    def writePhase(self, outputFile):
        data_list = []
        for ind in self:
            if ind.haplotypes.ndim == 2:  # diploid
                data_list.append((ind.idx, ind.haplotypes[0]))
                data_list.append((ind.idx, ind.haplotypes[1]))
            elif ind.haplotypes.ndim == 1:  # haploid
                data_list.append((ind.idx, ind.haplotypes))
                data_list.append((ind.idx, ind.haplotypes))

        MultiThreadIO.writeLines(outputFile, data_list, str)


    def writeDosages(self, outputFile):
        data_list = []
        for ind in self :
            if ind.dosages is not None:
                data_list.append( (ind.idx, ind.dosages) )
            else:
                dosages = ind.genotypes.copy()
                dosages[dosages == 9] = 1
                data_list.append( (ind.idx, dosages) )

        MultiThreadIO.writeLines(outputFile, data_list, "{:.4f}".format)


    def writeGenotypes_prefil(self, outputFile):
        # print("Output is using filled genotypes. Filling missing with a value of 1")
        # fillValues = np.full(1, self.nLoci)

        print("Output is using filled genotypes. Filling missing with rounded allele frequency")
        self.setMaf()
        fillValues = np.round(self.maf)

        with open(outputFile, 'w+') as f:
            for idx, ind in self.individuals.items():
                fill(ind.genotypes, fillValues)
                self.writeLine(f, ind.idx, ind.genotypes, str)


    def writeGenotypesPed(self, outputFile):
        """Write genotypes in PLINK plain text format"""
        data_list = []
        for ind in self:
            # Split genotypes into 'pseudo' haplotypes such that
            # the first allele/haplotype of a heterozygous locus is always 0
            missing = ind.genotypes == 9
            h1 = ind.genotypes//2
            h2 = ind.genotypes - h1
            h1[missing] = 9
            h2[missing] = 9
            alleles = self.encode_alleles(np.vstack([h1, h2]))
            data_list.append( (ind.idx, alleles) )
        MultiThreadIO.writeLinesPlinkPlainTxt(outputFile, data_list)


    def writePhasePed(self, outputFile):
        """Write phased data (i.e. haplotypes) in PLINK plain text .ped format"""
        data_list = []
        for ind in self:
            if ind.haplotypes.ndim == 2:  # diploid
                alleles = self.encode_alleles(ind.haplotypes)
            elif ind.haplotypes.ndim == 1:  # haploid
                diploid = np.vstack([ind.haplotypes, ind.haplotypes])
                alleles = self.encode_alleles(diploid)
            data_list.append( (ind.idx, alleles) )

        MultiThreadIO.writeLinesPlinkPlainTxt(outputFile, data_list)


    def writeLine(self, f, idx, data, func) :
        f.write(idx + ' ' + ' '.join(map(func, data)) + '\n')


@jit(nopython=True)
def fill(genotypes, fillValue):
    for i in range(len(genotypes)):
        if genotypes[i] == 9:
            genotypes[i] = fillValue[i]

@jit(nopython=True)
def addIfNotMissing(array1, counts, array2):
    for i in range(len(array1)):
        if array2[i] != 9:
            array1[i] += array2[i]
            counts[i] += 2


@jit(nopython=True)
def addIfMissing(array1, array2):
    for i in range(len(array1)):
        if array2[i] == 9:
            array1[i] += 1

@jit(nopython=True)
def doubleIfNotMissing(calledGenotypes):
    nLoci = len(calledGenotypes)
    for i in range(nLoci):
        if calledGenotypes[i] == 1:
            calledGenotypes[i] = 2

@jit(nopython=True)
def setMissing(calledGenotypes, matrix, thresh) :
    nLoci = len(calledGenotypes)
    for i in range(nLoci):
        if matrix[calledGenotypes[i],i] < thresh:
            calledGenotypes[i] = 9


















import numba
from numba import jit, float32, int8, int64, optional, boolean
from numba.experimental import jitclass
import numpy as np

def setup_individual(ind):
        fillInPhaseFromGenotypes(ind.haplotypes[0], ind.genotypes)
        fillInPhaseFromGenotypes(ind.haplotypes[1], ind.genotypes)
        fillInGenotypesFromPhase(ind.genotypes, ind.haplotypes[0], ind.haplotypes[1])

def align_individual(ind):
    # Note: The note below is no longer true.
    # Note: We never directly set genotypes so no need to go from genotypes -> phase

    # Fill in phase from genotypes.
    fillInPhaseFromGenotypes(ind.haplotypes[0], ind.genotypes)
    fillInPhaseFromGenotypes(ind.haplotypes[1], ind.genotypes)

    # Fill in genotypes from phase, and fill in haplotypes from each other.
    fillInGenotypesFromPhase(ind.genotypes, ind.haplotypes[0], ind.haplotypes[1])
    fillInCompPhase(ind.haplotypes[0], ind.genotypes, ind.haplotypes[1])
    fillInCompPhase(ind.haplotypes[1], ind.genotypes, ind.haplotypes[0])

def ind_fillInGenotypesFromPhase(ind):
    #Note: We never directly set genotypes so no need to go from genotypes -> phase
    fillInGenotypesFromPhase(ind.genotypes, ind.haplotypes[0], ind.haplotypes[1])

def fillFromParents(ind):
    if ind.sire is not None:
        if ind.sire.genotypes is not None:
            sirePhase = getPhaseFromGenotypes(ind.sire.genotypes)
            fillIfMissing(ind.haplotypes[0], sirePhase)
    
    if ind.dam is not None:
        if ind.dam.genotypes is not None:
            damPhase = getPhaseFromGenotypes(ind.dam.genotypes)
            fillIfMissing(ind.haplotypes[0], damPhase)

@jit(nopython=True)
def fillIfMissing(orig, new):
    for i in range(len(orig)):
        if orig[i] == 9:
            orig[i] = new[i]

@jit(nopython=True)
def fillInGenotypesFromPhase(geno, phase1, phase2):
    for i in range(len(geno)):
        if geno[i] == 9:
            if phase1[i] != 9 and phase2[i] != 9:
                geno[i] = phase1[i] + phase2[i]
@jit(nopython=True)
def fillInCompPhase(target, geno, compPhase):
    for i in range(len(geno)):
        if target[i] == 9:
            if geno[i] != 9:
                if compPhase[i] != 9:
                    target[i] = geno[i] - compPhase[i]
    
@jit(nopython=True)
def fillInPhaseFromGenotypes(phase, geno):
    for i in range(len(geno)):
        if phase[i] == 9 :
            if geno[i] == 0: phase[i] = 0
            if geno[i] == 2: phase[i] = 1

@jit(nopython=True)
def getPhaseFromGenotypes(geno):
    phase = np.full(len(geno), 9, dtype = np.int8)
    for i in range(len(geno)):
        if phase[i] == 9 :
            if geno[i] == 0: phase[i] = 0
            if geno[i] == 2: phase[i] = 1
    return phase


# def ind_randomlyPhaseRandomPoint(ind):
#     maxLength = len(ind.genotypes)
#     midpoint = np.random.normal(maxLength/2, maxLength/10)
#     while midpoint < 0 or midpoint > maxLength:
#         midpoint = np.random.normal(maxLength/2, maxLength/10)
#     midpoint = int(midpoint)
#     randomlyPhaseMidpoint(ind.genotypes, ind.haplotypes, midpoint)

# def ind_randomlyPhaseMidpoint(ind, midpoint= None):
#     if midpoint is None: midpoint = int(len(ind.genotypes)/2)
#     randomlyPhaseMidpoint(ind.genotypes, ind.haplotypes, midpoint)

# @jit(nopython=True)
# def randomlyPhaseMidpoint(geno, phase, midpoint):

#     index = 0
#     e = 1
#     changed = False
#     while not changed :
#         if geno[midpoint + index * e] == 1:
#             phase[0][midpoint + index * e] = 0
#             phase[1][midpoint + index * e] = 1
#             changed = True
#         e = -e
#         if e == -1: index += 1
#         if index >= midpoint: changed = True




import sys
import argparse
import re

import numpy as np
from numba import jit

import random
import warnings

alphaplinkpython_avail = False
try:
    import alphaplinkpython
    from alphaplinkpython import PlinkWriter
    alphaplinkpython_avail = True
except ImportError:
    alphaplinkpython_avail = False

##Global: 
args = None

def getParser(program) :
    parser = argparse.ArgumentParser(description='')
    core_parser = parser.add_argument_group("Core arguments")
    core_parser.add_argument('-out', required=True, type=str, help='The output file prefix.')

    addInputFileParser(parser)

    #Genotype files.
    # output_options_parser = parser.add_argument_group("Output options")

    # output_options_parser.add_argument('-writekey', default="id", required=False, type=str, help='Determines the order in which individuals are ordered in the output file based on their order in the corresponding input file. Animals not in the input file are placed at the end of the file and sorted in alphanumeric order. These animals can be surpressed with the "-onlykeyed" option. Options: id, pedigree, genotypes, sequence, segregation. Defualt: id.')
    # output_options_parser.add_argument('-onlykeyed', action='store_true', required=False, help='Flag to surpress the animals who are not present in the file used with -outputkey. Also surpresses "dummy" animals.')
    
    if program == "Default":
        pass

    if program == "AlphaImpute" :
        core_impute_parser = parser.add_argument_group("Impute options")
        core_impute_parser.add_argument('-no_impute', action='store_true', required=False, help='Flag to read in the files but not perform imputation.')
        core_impute_parser.add_argument('-no_phase', action='store_true', required=False, help='Flag to not do HD phasing initially.')
        core_impute_parser.add_argument('-maxthreads',default=1, required=False, type=int, help='Number of threads to use. Default: 1.')
        core_impute_parser.add_argument('-binaryoutput', action='store_true', required=False, help='Flag to write out the genotypes as a binary plink output.')

    if program in ["AlphaPeel", "AlphaMGS", "AlphaCall"]:
        probability_parser = parser.add_argument_group("Genotype probability arguments")
        add_arguments_from_dictionary(probability_parser, get_probability_options(), None)


        
    if program in ["longreads"]:
        longread_parser = parser.add_argument_group("Long read arguments")
        longread_parser.add_argument('-longreads', default=None, required=False, type=str, nargs="*", help='A read file.')

    if program == "AlphaPlantImpute" :
        core_plant_parser = parser.add_argument_group("Mandatory arguments")
        core_plant_parser.add_argument('-plantinfo', default=None, required=False, type=str, nargs="*", help='A plant info file.')

        
    if program == "AlphaMGS" :
        core_assign_parser = parser.add_argument_group("Core assignement arguments")
        core_assign_parser.add_argument('-potentialgrandsires', default=None, required=False, type=str, help='A list of potential dams for each individual.')
        core_assign_parser.add_argument('-usemaf', action='store_true', required=False, help='A flag to use the minor allele frequency when constructing genotype estimates for the sire and maternal grandsire. Not recomended for small input pedigrees.')

    if program == "AlphaCall":
        call_parser = parser.add_argument_group("AlphaCall arguments")
        call_parser.add_argument('-threshold', default=None, required=False, type=float, help='Genotype calling threshold. Use. .3 for best guess genotype.')
        call_parser.add_argument('-sexchrom', action='store_true', required=False, help='A flag to that this is a sex chromosome. Sex needs to be given in the pedigree file. This is currently an experimental option.')

    return parser


def addInputFileParser(parser):
    genotype_parser = parser.add_argument_group("Input arguments")
    add_arguments_from_dictionary(genotype_parser, get_input_options(), None)

    output_options_parser = parser.add_argument_group("Output options")
    add_arguments_from_dictionary(output_options_parser, get_output_options(), None)


def get_input_options():

    parse_dictionary = dict()
    parse_dictionary["bfile"] = lambda parser: parser.add_argument('-bfile',   default=None, required=False, type=str, nargs="*", help='A file in plink (binary) format. Only stable on Linux).')
    parse_dictionary["genotypes"] = lambda parser: parser.add_argument('-genotypes', default=None, required=False, type=str, nargs="*", help='A file in AlphaGenes format.')
    parse_dictionary["reference"] = lambda parser: parser.add_argument('-reference', default=None, required=False, type=str, nargs="*", help='A haplotype reference panel in AlphaGenes format.')
    parse_dictionary["seqfile"] = lambda parser: parser.add_argument('-seqfile', default=None, required=False, type=str, nargs="*", help='A sequence data file.')
    parse_dictionary["pedigree"] = lambda parser: parser.add_argument('-pedigree',default=None, required=False, type=str, nargs="*", help='A pedigree file in AlphaGenes format.')
    parse_dictionary["phasefile"] = lambda parser: parser.add_argument('-phasefile',default=None, required=False, type=str, nargs="*", help='A phase file in AlphaGenes format.')
    parse_dictionary["startsnp"] = lambda parser: parser.add_argument('-startsnp',default=None, required=False, type=int, help="The first marker to consider. The first marker in the file is marker '1'. Default: 1.")
    parse_dictionary["stopsnp"] = lambda parser: parser.add_argument('-stopsnp',default=None, required=False, type=int, help='The last marker to consider. Default: all markers considered.')
    parse_dictionary["seed"] = lambda parser: parser.add_argument('-seed',default=None, required=False, type=int, help='A random seed to use for debugging.')
    
    return parse_dictionary

def get_output_options():
    parse_dictionary = dict()

    parse_dictionary["writekey"] = lambda parser: parser.add_argument('-writekey', default="id", required=False, type=str, help='Determines the order in which individuals are ordered in the output file based on their order in the corresponding input file. Animals not in the input file are placed at the end of the file and sorted in alphanumeric order. These animals can be surpressed with the "-onlykeyed" option. Options: id, pedigree, genotypes, sequence, segregation. Defualt: id.')
    parse_dictionary["onlykeyed"] = lambda parser: parser.add_argument('-onlykeyed', action='store_true', required=False, help='Flag to surpress the animals who are not present in the file used with -outputkey. Also surpresses "dummy" animals.')
    parse_dictionary["iothreads"] = lambda parser: parser.add_argument('-iothreads', default=1, required=False, type=int, help='Number of threads to use for io. Default: 1.')

    return parse_dictionary

def get_multithread_options():
    parse_dictionary = dict()
    parse_dictionary["iothreads"] = lambda parser: parser.add_argument('-iothreads', default=1, required=False, type=int, help='Number of threads to use for input and output. Default: 1.')
    parse_dictionary["maxthreads"] = lambda parser: parser.add_argument('-maxthreads', default=1, required=False, type=int, help='Maximum number of threads to use for analysis. Default: 1.')
    return parse_dictionary

def get_probability_options():
    parse_dictionary = dict()
    parse_dictionary["error"] = lambda parser: parser.add_argument('-error', default=0.01, required=False, type=float, help='Genotyping error rate. Default: 0.01.')
    parse_dictionary["seqerror"] = lambda parser: parser.add_argument('-seqerror', default=0.001, required=False, type=float, help='Assumed sequencing error rate. Default: 0.001.')
    parse_dictionary["recombination"] = lambda parser: parser.add_argument('-recomb', default=1, required=False, type=float, help='Recombination rate per chromosome. Default: 1.')
    return parse_dictionary



def add_arguments_from_dictionary(parser, arg_dict, options = None):
    if options is None:
        for key, value in arg_dict.items():
            value(parser)
    else:
        for option in options:
            if option in arg_dict:
                arg_dict[option](parser)
            else:
                print("Option not found:", option, arg_dict)


def parseArgs(program, parser = None, no_args = False):
    global args
    args = rawParseArgs(program, parser, no_args = no_args)
    args.program = program

    # We want start/stop snp to be in python format (i.e. 0 to n-1).
    # Input values are between 1 to n.
    try:
        if args.startsnp is not None:
            args.startsnp -= 1
            args.stopsnp -= 1
        ##Add any necessary code to check args here.
    except AttributeError as error:
        pass

    return args

def rawParseArgs(program, parser = None, no_args = False) :
    if parser is None:
        parser = getParser(program)

    if no_args:
        return parser.parse_args(["-out", "out"])

    else:
        args = sys.argv[1:]
        if len(args) == 0 : 
            parser.print_help(sys.stderr)
            sys.exit(1)

        
        if len(args) == 1:
            with open(args[0]) as f:
                args = []
                for line in f:
                    if line[0] != "-": line = "-" + line
                    args.extend(re.split(r"[,|\s|\n]+",line))
        for arg in args:
            if len(arg) == 0:
                args.remove(arg)

        for i, arg in enumerate(args):
            if len(arg) > 0 and arg[0] == "-":
                args[i] = str.lower(arg)

        return parser.parse_args(args)


parseArgs("Default", parser = None, no_args = True)

@jit(nopython=True)
def setNumbaSeeds(seed):
    np.random.seed(seed)
    random.seed(seed)

def readInPedigreeFromInputs(pedigree, args, genotypes=True, haps=False, reads=False, update_coding=False):
    # Try catch is incase the program does not have a seed option.
    seed = getattr(args, "seed", None)

    if seed is not None:
        np.random.seed(args.seed)
        random.seed(args.seed)
        setNumbaSeeds(args.seed)


    startsnp = getattr(args, "startsnp", None)
    stopsnp = getattr(args, "stopsnp", None)

    pedigree.args = args
    pedigreeReadIn = False

    pedigree_args = getattr(args, "pedigree", None)   
    if pedigree_args is not None: 
        pedigreeReadIn = True
        for ped in args.pedigree:
            pedigree.readInPedigree(ped)

    # This gets the attribute from args, but returns None if the atribute is not valid.
    genotypes = getattr(args, "genotypes", None)    
    if genotypes is not None: 
        for geno in args.genotypes:
            pedigree.readInGenotypes(geno, startsnp, stopsnp)
    
    reference = getattr(args, "reference", None)    
    if reference is not None: 
        for ref in args.reference:
            pedigree.readInReferencePanel(ref, startsnp, stopsnp)
    
    seqfile = getattr(args, "seqfile", None)    
    if seqfile is not None: 
        for seq in args.seqfile:
            pedigree.readInSequence(seq, startsnp, stopsnp)
    
    phasefile = getattr(args, "phasefile", None)    
    if phasefile is not None: 
        if args.program == "AlphaPeel":
            print("Use of an external phase file is not currently supported. Phase information will be translated to genotype probabilities. If absolutely necessary use a penetrance file instead.") 
        for phase in args.phasefile:
            pedigree.readInPhase(phase, startsnp, stopsnp)
    
    bfile = getattr(args, "bfile", None)
    if bfile is not None: 
        global alphaplinkpython_avail
        if alphaplinkpython_avail:
            for plink in args.bfile:
                if pedigreeReadIn == True:
                    print(f"Pedigree file read in from -pedigree option. Reading in binary plink file {plink}. Pedigree information in the .fam file will be ignored.")
                readInGenotypesPlink(pedigree, plink, startsnp, stopsnp, pedigreeReadIn)
        else:
            warnings.warn("The module alphaplinkpython was not found. Plink files cannot be read in and will be ignored.")

    # Note: need to read .bim before .ped as the .bim sets the allele coding to use for reading the .ped
    bim = getattr(args, 'bim', None)
    if bim is not None:
        for file in args.bim:
            pedigree.readInBim(file, startsnp, stopsnp)

    ped = getattr(args, 'ped', None)
    if ped is not None:
        for file in args.ped:
            pedigree.readInPed(file, startsnp, stopsnp, haps=False, update_coding=update_coding)

    #It's important that these happen after all the datafiles are read in.
    #Each read in can add individuals. This information needs to be calculated on the final pedigree.
    pedigree.fillIn(genotypes, haps, reads)
    pedigree.setUpGenerations()
    pedigree.setupFamilies()

def readMapFile(mapFile, start = None, stop = None) :
    ids = []
    chrs = []
    positions = []
    with open(mapFile) as f:
        for line in f:
            parts = line.split(); 
            try:
                positions.append(float(parts[2]))
                chrs.append(parts[1])
                ids.append(parts[0])
            except ValueError:
                pass

    if start is None:
        start = 0
        stop = len(ids)

    return (ids[start:stop], chrs[start:stop], positions[start:stop])


# def getPositionsFromMap(mapFileName, nLoci) :
#     if mapFileName is None:
#         return np.arange(nLoci, dtype = np.float32)

#     positions = []
#     with open(mapFileName) as f:
#         for line in f:
#             parts = line.split(); 
#             try:
#                 positions.append(float(parts[1]))
#             except ValueError:
#                 pass
#     if len(positions) != nLoci :
#         raise ValueError(f"Number of loci not equal to map file length {nLoci}, {len(positions)}")

#     return np.array(positions, dtype = np.float32)


def readInSeg(pedigree, fileName, start=None, stop = None):
    print("Reading in seg file:", fileName)
    if start is None: start = 0
    if stop is None: stop = pedigree.nLoci
    nLoci = stop - start + 1 #Contains stop.
    


    print(pedigree.maxIdn)
    seg = np.full((pedigree.maxIdn, 4, nLoci), .25, dtype = np.float32)
    index = 0

    fileNColumns = 0

    indHit = np.full(pedigree.maxIdn, 0, dtype = np.int64)

    with open(fileName) as f:
        e = 0
        currentInd = None
        for line in f:
            parts = line.split();
            idx = parts[0]; 

            if fileNColumns == 0:
                fileNColumns = len(parts)
            if fileNColumns != len(parts):
                raise ValueError(f"The length of the line is not the expected length. Expected {fileNColumns} got {len(parts)} on individual {idx} and line {e}.")

            segLine=np.array([float(val) for val in parts[(start+1):(stop+2)]], dtype = np.float32)
            if len(segLine) != nLoci:
                raise ValueError(f"The length of the line subsection is not the expected length. Expected {nLoci} got {len(segLine)} on individual {idx} and line {e}.")

            if idx not in pedigree.individuals:
                print(f"Individual {idx} not found in pedigree. Individual ignored.")
            else:
                ind = pedigree.individuals[idx]
                if e == 0: 
                    currentInd = ind.idx
                if currentInd != ind.idx:
                    raise ValueError(f"Unexpected individual. Expecting individual {currentInd}, but got ind {ind.idx} on value {e}")
                seg[ind.idn,e,:] = segLine
                e = (e+1)%4
                ind.fileIndex['segregation'] = index; index += 1
                indHit[ind.idn] += 1
        for ind in pedigree:
            if indHit[ind.idn] != 4:
                print(f"No segregation information found for individual {ind.idx}")

    return seg


def writeIdnIndexedMatrix(pedigree, matrix, outputFile):
    np.set_printoptions(suppress=True)
    print("Writing to ", outputFile)
    with open(outputFile, 'w+') as f:

        for idx, ind in pedigree.writeOrder():
            if len(matrix.shape) == 2 :
                tmp = np.around(matrix[ind.idn, :], decimals = 4)
                f.write(' '.join(map(str, tmp)))
                # f.write('\n')
            if len(matrix.shape) == 3 :
                for i in range(matrix.shape[1]) :
                    f.write(idx + " ")
                    tmp2 = map("{:.4f}".format, matrix[ind.idn,i, :].tolist())
                    tmp3 = ' '.join(tmp2)
                    f.write(tmp3)
                    f.write('\n')
                # f.write('\n')
def writeFamIndexedMatrix(pedigree, matrix, outputFile):
    np.set_printoptions(suppress=True)
    print("Writing to ", outputFile)
    with open(outputFile, 'w+') as f:

        for fam in pedigree.getFamilies():
            if len(matrix.shape) == 2 :
                tmp = np.around(matrix[fam.idn, :], decimals = 4)
                f.write(' '.join(map(str, tmp)))
                # f.write('\n')
            if len(matrix.shape) == 3 :
                for i in range(matrix.shape[1]) :
                    f.write(str(fam.idn) + " ")
                    tmp2 = map("{:.4f}".format, matrix[fam.idn,i, :].tolist())
                    tmp3 = ' '.join(tmp2)
                    f.write(tmp3)
                    f.write('\n')
                # f.write('\n')

def writeOutGenotypesPlink(pedigree, fileName):

    global alphaplinkpython_avail
    if alphaplinkpython_avail:
        import alphaplinkpython
        from alphaplinkpython import PlinkWriter

        ped = [getFamString(ind) for ind in pedigree]

        nLoci = pedigree.nLoci
        nInd = len(pedigree.individuals)
        genotypes = np.full((nLoci, nInd), 0, dtype = np.int8)

        for i, ind in enumerate(pedigree):
            genotypes[:,i] = ind.genotypes

        genotypeIds = ["snp" + str(i+1) for i in range(nLoci)]
        genotypePos = [i + 1 for i in range(nLoci)]
        if args.startsnp is not None:
            genotypeIds = ["snp" + str(i + args.startsnp + 1) for i in range(nLoci)]
            genotypePos = [i + args.startsnp + 1 for i in range(nLoci)]

        PlinkWriter.writeFamFile(fileName + ".fam", ped)
        # PlinkWriter.writeBimFile(genotypeIds, fileName + ".bim")
        writeSimpleBim(genotypeIds, genotypePos, fileName + ".bim")
        PlinkWriter.writeBedFile(genotypes, fileName + ".bed")
    else:
        warnings.warn("The module alphaplinkpython was not found. Plink files cannot be written out and will be ignored.")

def writeSimpleBim(genotypeIds, genotypePos, fileName) :
    with open(fileName, "w") as file:
        for i in range(len(genotypeIds)):
            line = f"1 {genotypeIds[i]} 0 {genotypePos[i]} A B \n"
            file.write(line)



def readInGenotypesPlink(pedigree, fileName, startsnp, endsnp, externalPedigree = False):
    bim = PlinkWriter.readBimFile(fileName + '.bim')
    fam = PlinkWriter.readFamFile(fileName + '.fam')
    bed = PlinkWriter.readBedFile(fileName + '.bed', bim, fam)
    if startsnp is not None:
        bed = bed[startsnp:endsnp,:]

    pedList = [[line.getId(), line.getSire(), line.getDam()] for line in fam]
    idList = [line.getId() for line in fam]

    pedigree.readInFromPlink(idList, pedList, bed, externalPedigree)


def getFamString(ind):
    sireStr = 0
    damStr = 0
    if ind.sire is not None:
        sireStr = ind.sire.idx
    if ind.dam is not None:
        damStr = ind.dam.idx

    return [str(ind.idx), str(sireStr), str(damStr)]
    # return [str(ind.idx).encode('utf-8'), str(sireStr).encode('utf-8'), str(damStr).encode('utf-8')]

# @profile
# def writeIdnIndexedMatrix2(pedigree, matrix, outputFile):
#     np.set_printoptions(suppress=True)
#     print("Writing to ", outputFile)
#     with open(outputFile, 'w+') as f:

#         for idx, ind in pedigree.individuals.items():
#             if len(matrix.shape) == 2 :
#                 tmp = np.around(matrix[ind.idn, :], decimals = 4)
#                 f.write(' '.join(map(str, tmp)))
#                 f.write('\n')
#             if len(matrix.shape) == 3 :
#                 for i in range(matrix.shape[1]) :
#                     tmp2 = tuple(map("{:.4f}".format, matrix[ind.idn,i, :]))
#                     tmp3 = ' '.join(tmp2)
#                     f.write(tmp3)
#                     f.write('\n')


def print_boilerplate(name, version = None, commit = None, date = None):
    """Print software name, version and contact"""
    width = 42  # width of 'website' line
    print('-' * width)
    print(f'{name:^{width}}')  # centre aligned
    print('-' * width)
    if version is not None:
        print(f'Version: {version}')
    if commit is not None:
        print(f'Commit:  {commit}')
    if date is not None:
        print(f'Date:    {date}')
    
    print('Email:   alphagenes@roslin.ed.ac.uk')
    print('Website: http://alphagenes.roslin.ed.ac.uk')
    print('-' * width)
    